"""HDX Location Tests"""
from hdx.location.country import Country

from hdx.api.configuration import Configuration
from hdx.api.locations import Locations


class MyConfiguration:
    def call_remoteckan(self, a, b):
        return [{"name": "zaf", "title": "South Africa"}]


class TestHDXLocations:
    def test_validlocations(self, project_config_yaml):
        Country.countriesdata(use_live=False)
        validlocations = [{"name": "shn", "title": "St. Helena"}]
        assert (
            Locations.get_HDX_code_from_location(
                "sh", locations=validlocations
            )
            is None
        )
        assert Locations.get_HDX_code_from_location_partial(
            "sh", locations=validlocations
        ) == (None, False)
        assert (
            Locations.get_location_from_HDX_code(
                "shn", locations=validlocations
            )
            == "St. Helena"
        )
        validlocations = [
            {"name": "zmb", "title": "Zambia"},
            {"name": "pry", "title": "Paraguay"},
        ]
        Locations.set_validlocations(validlocations)
        assert Locations.validlocations() == validlocations
        assert Locations.get_HDX_code_from_location_partial("NOT") == (
            None,
            False,
        )
        assert Locations.get_location_from_HDX_code("pr") is None
        assert Locations.get_HDX_code_from_location("zmb") == "ZMB"
        assert Locations.get_HDX_code_from_location_partial("zmb") == (
            "ZMB",
            True,
        )
        assert Locations.get_HDX_code_from_location("Z") is None
        assert Locations.get_HDX_code_from_location_partial("Z") == (
            "ZMB",
            False,
        )
        assert Locations.get_HDX_code_from_location_partial("Zambia") == (
            "ZMB",
            True,
        )
        assert Locations.get_HDX_code_from_location_partial("ZAM") == (
            "ZMB",
            False,
        )
        assert (
            Locations.get_location_from_HDX_code(
                "zmb", locations=validlocations
            )
            == "Zambia"
        )
        validlocations = [{"name": "shn", "title": "St. Helena"}]
        assert (
            Locations.get_HDX_code_from_location(
                "sh", locations=validlocations
            )
            is None
        )
        assert Locations.get_HDX_code_from_location_partial(
            "sh", locations=validlocations
        ) == (None, False)
        assert (
            Locations.get_location_from_HDX_code(
                "shn", locations=validlocations
            )
            == "St. Helena"
        )
        Configuration.setup(MyConfiguration())
        Locations.set_validlocations(None)
        assert Locations.get_HDX_code_from_location("zaf") == "ZAF"
        assert Locations.get_HDX_code_from_location_partial("zaf") == (
            "ZAF",
            True,
        )
        assert Locations.get_location_from_HDX_code("zaf") == "South Africa"



"""Resource Tests"""
import copy
import json
import os
from datetime import datetime, timezone
from os import remove
from os.path import basename, join

import pytest
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.downloader import DownloadError

from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXError
from hdx.data.resource import Resource

from . import MockResponse, dataset_resultdict
from .test_resource_view import resource_view_list, resource_view_mocklist

resultdict = {
    "cache_last_updated": None,
    "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
    "webstore_last_updated": None,
    "datastore_active": None,
    "id": "de6549d8-268b-4dfe-adaf-a4ae5c8510d5",
    "size": None,
    "hash": "",
    "description": "My Resource",
    "format": "csv",
    "last_modified": None,
    "url_type": "api",
    "mimetype": None,
    "cache_url": None,
    "name": "MyResource1",
    "created": "2016-06-07T08:57:27.367939",
    "url": "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/test_data.csv",
    "webstore_url": None,
    "mimetype_inner": None,
    "position": 0,
    "revision_id": "43765383-1fce-471f-8166-d6c8660cc8a9",
    "resource_type": "api",
}

searchdict = {
    "count": 4,
    "results": [
        {
            "size": None,
            "description": "ACLED-All-Africa-File_20160101-to-20160903.xlsx",
            "revision_id": "796b639e-f217-4d0e-bedd-4e3b35da4461",
            "cache_last_updated": None,
            "created": "2016-06-22T12:41:02.857171",
            "name": "ACLED-All-Africa-File_20160101-to-date.xlsx",
            "last_modified": None,
            "url_type": "api",
            "format": "XLSX",
            "cache_url": None,
            "url": "http://www.acleddata.com/wp-content/uploads/2016/09/ACLED.slxs",
            "state": "active",
            "position": 0,
            "revision_last_updated": "2016-06-28T14:17:27.150541",
            "webstore_last_updated": None,
            "mimetype": None,
            "package_id": "45f53bde-544c-4a4a-9c6f-d609481b8719",
            "resource_type": "api",
            "id": "d6d1c367-1980-4dc7-ab20-c23f02d3b9e7",
            "mimetype_inner": None,
            "hash": "",
            "webstore_url": None,
        },
        {
            "size": None,
            "description": "",
            "revision_id": "5259be68-e72f-4c02-be8e-61d7cd594a9b",
            "cache_last_updated": None,
            "created": "2016-03-23T14:17:02.272572",
            "name": "ACLED-Version-6-All-Africa-1997-2015_csv_dyadic.zip",
            "last_modified": None,
            "url_type": "api",
            "format": "ZIP",
            "cache_url": None,
            "url": "http://www.acleddata.com/wp-content/uploads/2016/01/ACLED.zip",
            "state": "active",
            "position": 0,
            "webstore_last_updated": None,
            "mimetype": None,
            "package_id": "71d852e4-e41e-4320-a770-9fc2bb87fb64",
            "resource_type": "api",
            "id": "866b3b60-5b2a-4ca3-9b76-665870cc6d71",
            "mimetype_inner": None,
            "webstore_url": None,
            "hash": "",
            "originalHash": "97196323",
        },
        {
            "size": None,
            "description": "ACLED-All-Africa-File_20160101-to-20160903_csv.zip",
            "revision_id": "796b639e-f217-4d0e-bedd-4e3b35da4461",
            "cache_last_updated": None,
            "created": "2016-06-22T12:41:02.857194",
            "name": "ACLED-All-Africa-File_20160101-to-date_csv.zip",
            "last_modified": None,
            "url_type": "api",
            "format": "zipped csv",
            "cache_url": None,
            "url": "http://www.acleddata.com/wp-content/uploads/2016/09/ACLED.zip",
            "state": "active",
            "position": 1,
            "revision_last_updated": "2016-06-28T14:17:27.150541",
            "webstore_last_updated": None,
            "mimetype": None,
            "package_id": "45f53bde-544c-4a4a-9c6f-d609481b8719",
            "resource_type": "api",
            "id": "74b74ae1-df0c-4716-829f-4f939a046811",
            "mimetype_inner": None,
            "hash": "",
            "webstore_url": None,
        },
        {
            "size": None,
            "description": "",
            "revision_id": "5259be68-e72f-4c02-be8e-61d7cd594a9b",
            "cache_last_updated": None,
            "created": "2016-03-23T14:17:08.594232",
            "name": "ACLED-Version-6-All-Africa-1997-2015_dyadic-file.xlsx",
            "last_modified": None,
            "url_type": "api",
            "format": "XLSX",
            "cache_url": None,
            "url": "http://www.acleddata.com/wp-content/uploads/2016/01/ACLEDxlsx",
            "state": "active",
            "position": 1,
            "webstore_last_updated": None,
            "mimetype": None,
            "package_id": "71d852e4-e41e-4320-a770-9fc2bb87fb64",
            "resource_type": "api",
            "id": "e1e16f5c-2380-4a28-87b1-f5d644f248e5",
            "mimetype_inner": None,
            "webstore_url": None,
            "hash": "",
            "originalHash": "97196323",
        },
    ],
}


def mockshow(url, datadict):
    if "show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}',
        )
    result = json.dumps(resultdict)
    if datadict["id"] == "74b74ae1-df0c-4716-829f-4f939a046811":
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}'
            % result,
        )
    if datadict["id"] == "74b74ae1-df0c-4716-829f-4f939a046812":
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}',
        )
    if datadict["id"] == "74b74ae1-df0c-4716-829f-4f939a046813":
        return MockResponse(
            200,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}',
        )
    if datadict["id"] == "74b74ae1-df0c-4716-829f-4f939a046814":
        resdictcopy = copy.deepcopy(resultdict)
        resdictcopy["url"] = "lalalala"
        result = json.dumps(resdictcopy)
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}'
            % result,
        )
    if datadict["id"] == "74b74ae1-df0c-4716-829f-4f939a046815":
        resdictcopy = copy.deepcopy(resultdict)
        resdictcopy["id"] = "datastore_unknown_resource"
        result = json.dumps(resdictcopy)
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}'
            % result,
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_show"}',
    )


def mockpatch(url, datadict):
    if "patch" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not patch", "__type": "TEST ERROR: Not Patch Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_patch"}',
        )
    result = json.dumps(resultdict)
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_patch"}'
        % result,
    )


def mockdataset(url, datadict):
    if "show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}',
        )
    result = json.dumps(dataset_resultdict)
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_show"}'
        % result,
    )


def mocksearch(url, datadict):
    if "search" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not search", "__type": "TEST ERROR: Not Search Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_search"}',
        )
    result = json.dumps(searchdict)
    if datadict["query"] == "name:ACLED":
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_search"}'
            % result,
        )
    if datadict["query"] == "fail":
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Validation Error", "__type": "Validation Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_search"}',
        )
    if datadict["query"] == "name:ajyhgr":
        return MockResponse(
            200,
            '{"success": true, "result": {"count": 0, "results": []}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_search"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_search"}',
    )


def mockresourceview(url, decodedata):
    if "delete" in url:
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_delete"}'
            % decodedata,
        )
    datadict = json.loads(decodedata)
    if "show" in url:
        if id not in datadict:
            return MockResponse(
                404,
                '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_show"}',
            )
        if datadict["title"] == "Data Explorer":
            result = json.dumps(resource_view_list[0])
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_show"}'
                % result,
            )
        if datadict["title"] == "Quick Charts":
            result = json.dumps(resource_view_list[1])
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_show"}'
                % result,
            )
    if "list" in url:
        return resource_view_mocklist(url, datadict)

    if "create" in url or "update" in url:
        if datadict["title"] == "Data Explorer":
            result = json.dumps(resource_view_list[0])
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_create"}'
                % result,
            )
        if datadict["title"] == "Quick Charts":
            result = json.dumps(resource_view_list[1])
            return MockResponse(
                200,
                '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_show"}'
                % result,
            )
    if "reorder" in url:
        result = json.dumps(
            {
                "id": "25982d1c-f45a-45e1-b14e-87d367413045",
                "order": [
                    "c06b5a0d-1d41-4a74-a196-41c251c76023",
                    "d80301b5-4abd-49bd-bf94-fa4af7b6e7a4",
                ],
            }
        )
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_reorder"}'
            % result,
        )

    return MockResponse(
        404,
        '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_view_show"}',
    )


class TestResource:
    resource_data = {
        "name": "MyResource1",
        "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
        "format": "xlsx",
        "url": "http://test/spreadsheet.xlsx",
        "description": "My Resource",
        "api_type": "api",
        "resource_type": "api",
    }

    datastore = None

    @pytest.fixture(scope="class")
    def static_yaml(self):
        return join("tests", "fixtures", "config", "hdx_resource_static.yml")

    @pytest.fixture(scope="class")
    def static_json(self):
        return join("tests", "fixtures", "config", "hdx_resource_static.json")

    @pytest.fixture(scope="class")
    def topline_yaml(self):
        return join(
            "tests", "fixtures", "config", "hdx_datasource_topline.yml"
        )

    @pytest.fixture(scope="class")
    def topline_json(self):
        return join(
            "tests", "fixtures", "config", "hdx_datasource_topline.json"
        )

    @pytest.fixture(scope="function")
    def read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_create(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return mockshow(url, datadict)
                if "resource_id" in datadict:
                    if (
                        datadict["resource_id"]
                        == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
                    ):
                        return MockResponse(
                            200,
                            '{"success": true, "result": {"fields": [{"type": "text", "id": "code"}, {"type": "text", "id": "title"}, {"type": "float", "id": "value"}, {"type": "timestamp", "id": "latest_date"}, {"type": "text", "id": "source"}, {"type": "text", "id": "source_link"}, {"type": "text", "id": "notes"}, {"type": "text", "id": "explore"}, {"type": "text", "id": "units"}], "method": "insert", "primary_key": "code", "resource_id": "bfa6b55f-10b6-4ba2-8470-33bb9a5194a5"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_create"}',
                        )
                if "create" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}',
                    )
                if datadict["name"] == "MyResource1":
                    resultdictcopy = copy.deepcopy(resultdict)
                    if files:
                        resultdictcopy["url_type"] = "upload"
                        resultdictcopy["resource_type"] = "file.upload"
                        filename = os.path.basename(files["upload"].name)
                        resultdictcopy[
                            "url"
                        ] = f"http://test-data.humdata.org/dataset/6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d/resource/de6549d8-268b-4dfe-adaf-a4ae5c8510d5/download/{filename}"
                    resultdictcopy["state"] = datadict["state"]

                    result = json.dumps(resultdictcopy)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}'
                        % result,
                    )
                if datadict["name"] == "MyResource2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}',
                    )
                if datadict["name"] == "MyResource3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}',
                    )
                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_create"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return mockshow(url, datadict)
                if "resource_id" in datadict:
                    if (
                        datadict["resource_id"]
                        == "74b74ae1-df0c-4716-829f-4f939a046811"
                    ):
                        return MockResponse(
                            200,
                            '{"success": true, "result": {"fields": [{"type": "text", "id": "code"}, {"type": "text", "id": "title"}, {"type": "float", "id": "value"}, {"type": "timestamp", "id": "latest_date"}, {"type": "text", "id": "source"}, {"type": "text", "id": "source_link"}, {"type": "text", "id": "notes"}, {"type": "text", "id": "explore"}, {"type": "text", "id": "units"}], "method": "insert", "primary_key": "code", "resource_id": "bfa6b55f-10b6-4ba2-8470-33bb9a5194a5"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_create"}',
                        )
                if "update" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not update", "__type": "TEST ERROR: Not Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}',
                    )
                if datadict["name"] == "MyResource1":
                    resultdictcopy = copy.deepcopy(resultdict)
                    merge_two_dictionaries(resultdictcopy, datadict)
                    if files:
                        resultdictcopy["url_type"] = "upload"
                        resultdictcopy["resource_type"] = "file.upload"
                        filename = os.path.basename(files["upload"].name)
                        resultdictcopy[
                            "url"
                        ] = f"http://test-data.humdata.org/dataset/6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d/resource/de6549d8-268b-4dfe-adaf-a4ae5c8510d5/download/{filename}"
                    result = json.dumps(resultdictcopy)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}'
                        % result,
                    )
                if datadict["name"] == "MyResource2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}',
                    )
                if datadict["name"] == "MyResource3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_update"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_delete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return mockshow(url, datadict)
                if "delete" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not delete", "__type": "TEST ERROR: Not Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_delete"}',
                    )
                if datadict["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_delete"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_datastore(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return mockshow(url, datadict)
                if (
                    "create" not in url
                    and "insert" not in url
                    and "upsert" not in url
                    and "delete" not in url
                    and "search" not in url
                ):
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create or delete", "__type": "TEST ERROR: Not Create or Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_action"}',
                    )
                if (
                    "delete" in url
                    and datadict["resource_id"] == "datastore_unknown_resource"
                ):
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_delete"}',
                    )
                if (
                    "delete" in url
                    and datadict["resource_id"]
                    == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
                ):
                    TestResource.datastore = "delete"
                    return MockResponse(
                        200,
                        '{"success": true, "result": {"resource_id": "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_delete"}',
                    )
                if (
                    "create" in url
                    and datadict["resource_id"] == "datastore_unknown_resource"
                ):
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_create"}',
                    )
                if (
                    "search" in url
                    and datadict["resource_id"] == "datastore_unknown_resource"
                ):
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_create"}',
                    )
                if (
                    "search" in url
                    and datadict["resource_id"] == "_table_metadata"
                ):
                    return MockResponse(
                        200,
                        '{"success": true, "result": {"include_total": true, "resource_id": "_table_metadata", "fields": [{"type": "int", "id": "_id"}, {"type": "name", "id": "name"}, {"type": "oid", "id": "oid"}, {"type": "name", "id": "alias_of"}], "records_format": "objects", "records": [{"_id":"f9cd60f3d7f2f6d0","name":"f9228459-d808-4b51-948f-68a5850abfde","oid":"919290","alias_of":null},{"_id":"7ae63490de9b7d7b","name":"af618a0b-09b8-42c8-836f-2be597e1ea34","oid":"135294","alias_of":null},{"_id":"1dc37f4e89988644","name":"748b40dd-7bd3-40a3-941b-e76f0bfbe0eb","oid":"117144","alias_of":null},{"_id":"2a554a61bd366206","name":"91c78d24-eab3-40b5-ba91-6b29bcda7178","oid":"116963","alias_of":null},{"_id":"fd787575143afe90","name":"9320cfce-4620-489a-bcbe-25c73867d4fc","oid":"107430","alias_of":null},{"_id":"a70093abd230f647","name":"b9d2eb36-e65c-417a-bc28-f4dadb149302","oid":"107409","alias_of":null},{"_id":"95fbdd2d06c07aea","name":"ca6a0891-8395-4d58-9168-6c44e17e0193","oid":"107385","alias_of":null}], "limit": 10000, "_links": {"start": "/api/action/datastore_search?limit=10000&resource_id=_table_metadata", "next": "/api/action/datastore_search?offset=10000&limit=10000&resource_id=_table_metadata"}, "total": 7}}',
                    )
                if (
                    "create" in url
                    or "insert" in url
                    or "upsert" in url
                    or "search" in url
                ) and datadict[
                    "resource_id"
                ] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5":
                    TestResource.datastore = "create"
                    return MockResponse(
                        200,
                        '{"success": true, "result": {"fields": [{"type": "text", "id": "code"}, {"type": "text", "id": "title"}, {"type": "float", "id": "value"}, {"type": "timestamp", "id": "latest_date"}, {"type": "text", "id": "source"}, {"type": "text", "id": "source_link"}, {"type": "text", "id": "notes"}, {"type": "text", "id": "explore"}, {"type": "text", "id": "units"}], "method": "insert", "primary_key": "code", "resource_id": "bfa6b55f-10b6-4ba2-8470-33bb9a5194a5"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=datastore_create"}',
                    )
                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=resource_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_patch(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return mockshow(url, datadict)
                return mockpatch(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_dataset(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mockdataset(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def search(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mocksearch(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_resourceview(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                return mockresourceview(url, decodedata)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_broken(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                if isinstance(data, dict):
                    datadict = {
                        k.decode("utf8"): v.decode("utf8")
                        for k, v in data.items()
                    }
                else:
                    datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return mockshow(url, datadict)
                if "broken" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not mark broken", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=hdx_mark_broken_link_in_resource"}',
                    )
                resultdictcopy = copy.deepcopy(resultdict)
                resultdictcopy["broken_link"] = True
                result = json.dumps(resultdictcopy)
                return MockResponse(
                    200,
                    '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=hdx_mark_broken_link_in_resource"}'
                    % result,
                )

        Configuration.read().remoteckan().session = MockSession()

    def test_read_from_hdx(self, configuration, read):
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        assert resource["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        assert resource["name"] == "MyResource1"
        assert resource["package_id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046812"
        )
        assert resource is None
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046813"
        )
        assert resource is None
        with pytest.raises(HDXError):
            Resource.read_from_hdx("ABC")

    def test_check_url_filetoupload(self, configuration):
        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        resource.check_url_filetoupload()
        resource.set_file_to_upload("abc")
        resource.check_url_filetoupload()
        resource["url"] = "lala"
        with pytest.raises(HDXError):
            resource.check_url_filetoupload()
        resource = Resource(resource_data)
        resource["format"] = "NOTEXIST"
        with pytest.raises(HDXError):
            resource.check_url_filetoupload()
        with pytest.raises(HDXError):
            resource.set_file_type("NOTEXIST")
        del resource["format"]
        with pytest.raises(HDXError):
            resource.check_url_filetoupload()

    def test_get_set_date_of_resource(self):
        resource = Resource(
            {"daterange_for_data": "[2020-01-07T00:00:00 TO *]"}
        )
        result = resource.get_date_of_resource(today=datetime(2020, 11, 17))
        assert result == {
            "startdate": datetime(2020, 1, 7, 0, 0, tzinfo=timezone.utc),
            "enddate": datetime(2020, 11, 17, 23, 59, 59, tzinfo=timezone.utc),
            "startdate_str": "2020-01-07T00:00:00+00:00",
            "enddate_str": "2020-11-17T23:59:59+00:00",
            "ongoing": True,
        }
        resource.set_date_of_resource("2020-02-09", "2020-10-20")
        result = resource.get_date_of_resource("%d/%m/%Y")
        assert result == {
            "startdate": datetime(2020, 2, 9, 0, 0, tzinfo=timezone.utc),
            "enddate": datetime(2020, 10, 20, 23, 59, 59, tzinfo=timezone.utc),
            "startdate_str": "09/02/2020",
            "enddate_str": "20/10/2020",
            "ongoing": False,
        }

    def test_check_required_fields(self, configuration):
        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        resource.check_url_filetoupload()
        resource.check_required_fields()

    def test_create_in_hdx(self, configuration, post_create):
        resource = Resource()
        with pytest.raises(HDXError):
            resource.create_in_hdx()
        resource["id"] = "74b74ae1-df0c-4716-829f-4f939a046811"
        resource["name"] = "LALA"
        with pytest.raises(HDXError):
            resource.create_in_hdx()

        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        resource.create_in_hdx()
        assert resource["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        assert resource["url_type"] == "api"
        assert resource["resource_type"] == "api"
        assert (
            resource["url"]
            == "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/test_data.csv"
        )

        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        filetoupload = join("tests", "fixtures", "test_data.csv")
        resource.set_file_to_upload(filetoupload)
        assert resource.get_file_to_upload() == filetoupload
        resource.create_in_hdx()
        assert resource["url_type"] == "upload"
        assert resource["resource_type"] == "file.upload"
        assert (
            resource["url"]
            == "http://test-data.humdata.org/dataset/6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d/resource/de6549d8-268b-4dfe-adaf-a4ae5c8510d5/download/test_data.csv"
        )
        assert resource["state"] == "active"
        resource.set_file_to_upload(filetoupload)
        resource.create_in_hdx()
        assert (
            resource["url"]
            == "http://test-data.humdata.org/dataset/6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d/resource/de6549d8-268b-4dfe-adaf-a4ae5c8510d5/download/test_data.csv"
        )
        resource_data["name"] = "MyResource2"
        resource = Resource(resource_data)
        with pytest.raises(HDXError):
            resource.create_in_hdx()

        resource_data["name"] = "MyResource3"
        resource = Resource(resource_data)
        with pytest.raises(HDXError):
            resource.create_in_hdx()

    def test_update_in_hdx(self, configuration, post_update):
        resource = Resource()
        resource["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            resource.update_in_hdx()
        resource["name"] = "LALA"
        with pytest.raises(HDXError):
            resource.update_in_hdx()

        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        assert resource["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        assert resource.get_file_type() == "csv"

        resource.set_file_type("XLSX")
        resource["id"] = "74b74ae1-df0c-4716-829f-4f939a046811"
        resource["name"] = "MyResource1"
        resource.update_in_hdx()
        assert resource["id"] == "74b74ae1-df0c-4716-829f-4f939a046811"
        assert resource["format"] == "xlsx"
        resource.set_file_type(".xsl")
        assert resource.get_file_type() == "xls"
        resource.set_file_type("XLSX")
        assert resource.get_file_type() == "xlsx"
        assert resource["url_type"] == "api"
        assert resource["resource_type"] == "api"
        assert (
            resource["url"]
            == "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/test_data.csv"
        )
        assert resource["state"] == "active"

        filetoupload = join("tests", "fixtures", "test_data.csv")
        resource.set_file_to_upload(
            filetoupload, guess_format_from_suffix=True
        )
        assert resource["format"] == "csv"
        resource.update_in_hdx()
        assert resource["url_type"] == "upload"
        assert resource["resource_type"] == "file.upload"
        assert (
            resource["url"]
            == "http://test-data.humdata.org/dataset/6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d/resource/de6549d8-268b-4dfe-adaf-a4ae5c8510d5/download/test_data.csv"
        )
        assert resource["state"] == "active"

        resource["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            resource.update_in_hdx()

        del resource["id"]
        with pytest.raises(HDXError):
            resource.update_in_hdx()

        resource.data = dict()
        with pytest.raises(HDXError):
            resource.update_in_hdx()

        resource_data = copy.deepcopy(TestResource.resource_data)
        resource_data["name"] = "MyResource1"
        resource_data["id"] = "74b74ae1-df0c-4716-829f-4f939a046811"
        resource = Resource(resource_data)
        resource.create_in_hdx()
        assert resource["id"] == "74b74ae1-df0c-4716-829f-4f939a046811"
        assert resource.get_file_type() == "xlsx"
        assert resource["state"] == "active"
        resource["format"] = "Geoservice"
        resource.update_in_hdx()
        assert resource.get_file_type() == "geoservice"
        resource["format"] = "NOTEXIST"
        with pytest.raises(HDXError):
            resource.update_in_hdx()

    def test_delete_from_hdx(self, configuration, post_delete):
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        resource.delete_from_hdx()
        del resource["id"]
        with pytest.raises(HDXError):
            resource.delete_from_hdx()

    def test_update_yaml(self, configuration, static_yaml):
        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        assert resource["name"] == "MyResource1"
        assert resource.get_file_type() == "xlsx"
        resource.update_from_yaml(static_yaml)
        assert resource["name"] == "MyResource1"
        assert resource.get_file_type() == "csv"

    def test_update_json(self, configuration, static_json):
        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        assert resource["name"] == "MyResource1"
        assert resource.get_file_type() == "xlsx"
        resource.update_from_json(static_json)
        assert resource["name"] == "MyResource1"
        assert resource.get_file_type() == "zipped csv"

    def test_patch(self, configuration, post_patch):
        resource = Resource()
        resource["id"] = "74b74ae1-df0c-4716-829f-4f939a046811"
        resource.update_in_hdx(
            operation="patch", batch_mode="KEEP_OLD", skip_validation=True
        )
        assert resource["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"

    def test_get_dataset(self, configuration, post_dataset):
        resource_data = copy.deepcopy(TestResource.resource_data)
        resource = Resource(resource_data)
        dataset = resource.get_dataset()
        assert dataset["id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        del resource["package_id"]
        with pytest.raises(HDXError):
            resource.get_dataset()

    def test_search_in_hdx(self, configuration, search):
        resources = Resource.search_in_hdx("name:ACLED")
        assert len(resources) == 4
        resources = Resource.search_in_hdx("name:ajyhgr")
        assert len(resources) == 0
        with pytest.raises(HDXError):
            Resource.search_in_hdx("fail")

    def test_download(self, configuration, read):
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        resource2 = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046814"
        )
        url, path = resource.download()
        remove(path)
        assert (
            url
            == "https://raw.githubusercontent.com/OCHA-DAP/hdx-python-api/master/tests/fixtures/test_data.csv"
        )
        assert basename(path) == "MyResource1.csv"
        resource["url"] = ""
        with pytest.raises(HDXError):
            resource.download()
        with pytest.raises(DownloadError):
            resource2.download()

    def test_datastore(
        self, configuration, post_datastore, topline_yaml, topline_json
    ):
        resource_ids = Resource.get_all_resource_ids_in_datastore()
        assert resource_ids == [
            "f9228459-d808-4b51-948f-68a5850abfde",
            "af618a0b-09b8-42c8-836f-2be597e1ea34",
            "748b40dd-7bd3-40a3-941b-e76f0bfbe0eb",
            "91c78d24-eab3-40b5-ba91-6b29bcda7178",
            "9320cfce-4620-489a-bcbe-25c73867d4fc",
            "b9d2eb36-e65c-417a-bc28-f4dadb149302",
            "ca6a0891-8395-4d58-9168-6c44e17e0193",
        ]
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        resource2 = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046815"
        )
        TestResource.datastore = None
        assert resource.has_datastore() is True
        assert TestResource.datastore == "create"
        TestResource.datastore = None
        assert resource2.has_datastore() is False

    def test_resource_views(self, configuration, post_resourceview):
        resource = Resource({"id": "25982d1c-f45a-45e1-b14e-87d367413045"})
        with pytest.raises(HDXError):
            resource.add_update_resource_view("123")
        resource_view = copy.deepcopy(resource_view_list[0])
        del resource_view["id"]
        del resource_view["package_id"]
        resource.add_update_resource_view(resource_view)
        resource_view = copy.deepcopy(resource_view_list[1])
        del resource_view["id"]
        del resource_view["package_id"]
        with pytest.raises(HDXError):
            resource.add_update_resource_views("123")
        resource.add_update_resource_views([resource_view])
        resource_views = resource.get_resource_views()
        assert (
            resource_views[0]["id"] == "d80301b5-4abd-49bd-bf94-fa4af7b6e7a4"
        )
        assert (
            resource_views[1]["id"] == "c06b5a0d-1d41-4a74-a196-41c251c76023"
        )
        with pytest.raises(HDXError):
            resource.delete_resource_view("123")
        resource.delete_resource_view("d80301b5-4abd-49bd-bf94-fa4af7b6e7a4")
        resource.delete_resource_view(resource_view)
        resource_view["title"] = "XXX"
        with pytest.raises(HDXError):
            resource.delete_resource_view(resource_view)
        with pytest.raises(HDXError):
            resource.reorder_resource_views("123")
        resource.reorder_resource_views(
            [
                "c06b5a0d-1d41-4a74-a196-41c251c76023",
                "d80301b5-4abd-49bd-bf94-fa4af7b6e7a4",
            ]
        )
        resource.reorder_resource_views(resource_view_list)
        resource_view = copy.deepcopy(resource_view_list[0])
        resource_view["id"] = "123"
        with pytest.raises(HDXError):
            resource.reorder_resource_views(
                [resource_view_list[1], resource_view]
            )

    def test_mark_broken(self, configuration, post_broken):
        resource = Resource.read_from_hdx(
            "74b74ae1-df0c-4716-829f-4f939a046811"
        )
        assert resource["id"] == "de6549d8-268b-4dfe-adaf-a4ae5c8510d5"
        assert resource["name"] == "MyResource1"
        assert resource["package_id"] == "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d"
        resource.mark_broken()
        assert resource.is_broken() is True



"""Facade to simplify project setup that calls project main function with kwargs"""
import logging
from typing import Any, Callable, Optional

import defopt
from hdx.utilities.easy_logging import setup_logging
from hdx.utilities.useragent import UserAgent

from hdx.api import __version__
from hdx.api.configuration import Configuration

logger = logging.getLogger(__name__)
setup_logging(log_file="errors.log")


def _create_configuration(
    user_agent: Optional[str] = None,
    user_agent_config_yaml: Optional[str] = None,
    user_agent_lookup: Optional[str] = None,
    hdx_url: Optional[str] = None,
    hdx_site: Optional[str] = None,
    hdx_read_only: bool = False,
    hdx_key: Optional[str] = None,
    hdx_config_json: Optional[str] = None,
    hdx_config_yaml: Optional[str] = None,
    project_config_json: Optional[str] = None,
    project_config_yaml: Optional[str] = None,
    hdx_base_config_json: Optional[str] = None,
    hdx_base_config_yaml: Optional[str] = None,
) -> str:
    """
    Create HDX configuration

    Args:
        user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
        user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
        user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.
        hdx_url (Optional[str]): HDX url to use. Overrides hdx_site.
        hdx_site (Optional[str]): HDX site to use eg. prod, test.
        hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
        hdx_key (Optional[str]): Your HDX key. Ignored if hdx_read_only = True.
        hdx_config_json (Optional[str]): Path to JSON HDX configuration OR
        hdx_config_yaml (Optional[str]): Path to YAML HDX configuration
        project_config_json (Optional[str]): Path to JSON Project configuration OR
        project_config_yaml (Optional[str]): Path to YAML Project configuration
        hdx_base_config_json (Optional[str]): Path to JSON HDX base configuration OR
        hdx_base_config_yaml (Optional[str]): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.

    Returns:
        str: HDX site url

    """
    arguments = locals()
    return Configuration._create(**arguments)


def facade(projectmainfn: Callable[[Any], None], **kwargs: Any):
    """Facade to simplify project setup that calls project main function. It infers
    command line arguments from the passed in function using defopt. The function passed
    in should have type hints and a docstring from which to infer the command line
    arguments. Any **kwargs given will be merged with command line arguments, with the
    command line arguments taking precedence.

    Args:
        projectmainfn ((Any) -> None): main function of project
        **kwargs: Configuration parameters to pass to HDX Configuration & other parameters to pass to main function

    Returns:
        None
    """

    #
    # Setting up configuration
    #
    func, argv = defopt.bind_known(projectmainfn, cli_options="all")
    for key in kwargs:
        name = f"--{key.replace('_', '-')}"
        if name not in argv:
            argv.append(name)
            argv.append(kwargs[key])
    site_url = defopt.run(_create_configuration, argv=argv, cli_options="all")

    logger.info("--------------------------------------------------")
    logger.info(f"> Using HDX Python API Library {__version__}")
    logger.info(f"> HDX Site: {site_url}")

    UserAgent.user_agent = Configuration.read().user_agent
    func()



"""Showcase Tests"""
import copy
import json
from os.path import join

import pytest
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.loader import load_yaml

from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXError
from hdx.data.showcase import Showcase

from . import MockResponse
from .test_vocabulary import vocabulary_mockshow

showcase_resultdict = {
    "relationships_as_object": [],
    "num_tags": 2,
    "id": "05e392bf-04e0-4ca6-848c-4e87bba10746",
    "metadata_created": "2017-07-03T07:50:49.474517",
    "metadata_modified": "2017-07-03T08:12:43.726624",
    "author": "Dan Mihaila",
    "author_email": "dan@gmail.com",
    "state": "active",
    "creator_user_id": "060468e4-2f33-4488-8504-c4b10cc34821",
    "type": "showcase",
    "resources": [],
    "tags": [
        {
            "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            "state": "active",
            "display_name": "economics",
            "id": "e1ff4b78-efcf-40e7-888b-7342169068e1",
            "name": "economics",
        },
        {
            "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            "state": "active",
            "display_name": "health",
            "id": "8a43b269-78cb-4a85-a4ce-7e3e928bc487",
            "name": "health",
        },
    ],
    "groups": [],
    "relationships_as_subject": [],
    "total_res_downloads": 0,
    "num_datasets": 2,
    "showcase_notes_formatted": "<p>lalala</p>",
    "image_url": "visual.png",
    "revision_id": "cc64364b-ede2-400a-bb9f-8e585a4f6399",
    "notes": "My Showcase",
    "url": "http://visualisation/url/",
    "title": "MyShowcase1",
    "image_display_url": "http://myvisual/visual.png",
    "name": "showcase-1",
}

datasetsdict = load_yaml(
    join("tests", "fixtures", "dataset_search_results.yml")
)
allsearchdict = load_yaml(
    join("tests", "fixtures", "showcase_all_search_results.yml")
)


def mockshow(url, datadict):
    if "package_list" in url:
        result = json.dumps(datasetsdict["results"])
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_package_list"}'
            % result,
        )
    if "_show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_show"}',
        )
    result = json.dumps(showcase_resultdict)
    if (
        datadict["id"] == "05e392bf-04e0-4ca6-848c-4e87bba10746"
        or datadict["id"] == "TEST1"
    ):
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_show"}'
            % result,
        )
    if datadict["id"] == "TEST2":
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_show"}',
        )
    if datadict["id"] == "TEST3":
        return MockResponse(
            200,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_show"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_show"}',
    )


def mockallsearch(url, datadict):
    if "search" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not search", "__type": "TEST ERROR: Not Search Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    if "metadata_modified" in datadict["fq"]:
        newsearchdict = copy.deepcopy(allsearchdict)
        newsearchdict["results"] = [newsearchdict["results"][0]]
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
            % json.dumps(newsearchdict),
        )
    if datadict["q"] == "ACLED":
        newsearchdict = copy.deepcopy(allsearchdict)
        if datadict["rows"] == 11:
            newsearchdict["results"].append(newsearchdict["results"][0])
        elif datadict["rows"] == 6:
            if datadict["start"] == 2:
                newsearchdict["results"] = newsearchdict["results"][2:8]
        elif datadict["rows"] == 5:
            if datadict["start"] == 0:
                newsearchdict["count"] = 5
                newsearchdict["results"] = newsearchdict["results"][:5]
            elif datadict["start"] == 5:
                newsearchdict["count"] = 6  # return wrong count
                newsearchdict["results"] = newsearchdict["results"][:5]
            else:
                newsearchdict["count"] = 0
                newsearchdict["results"] = list()
        result = json.dumps(newsearchdict)
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
            % result,
        )
    if datadict["q"] == "*:*":
        newsearchdict = copy.deepcopy(allsearchdict)
        newsearchdict["results"].extend(
            copy.deepcopy(newsearchdict["results"])
        )
        for i, x in enumerate(newsearchdict["results"]):
            x["id"] = f'{x["id"]}{i}'
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
            % json.dumps(newsearchdict),
        )
    if datadict["q"] == '"':
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Validation Error", "__type": "Validation Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    if datadict["q"] == "ajyhgr":
        return MockResponse(
            200,
            '{"success": true, "result": {"count": 0, "results": []}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
    )


class TestShowcase:
    showcase_data = {
        "title": "MyShowcase1",
        "notes": "My Showcase",
        "package_id": "6f36a41c-f126-4b18-aaaf-6c2ddfbc5d4d",
        "image_url": "http://myvisual/visual.png",
        "name": "showcase-1",
        "url": "http://visualisation/url/",
        "tags": [{"name": "economy"}, {"name": "health"}],
        "dataset_ids": ["a89c6260-6392-416e-bcbc-eb2c5f1d7add"],
    }
    association = None

    @pytest.fixture(scope="class")
    def static_yaml(self):
        return join("tests", "fixtures", "config", "hdx_showcase_static.yml")

    @pytest.fixture(scope="class")
    def static_json(self):
        return join("tests", "fixtures", "config", "hdx_showcase_static.json")

    @pytest.fixture(scope="function")
    def read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "association_delete" in url:
                    TestShowcase.association = "delete"
                    return MockResponse(
                        200,
                        '{"success": true, "result": null, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_package_association_delete"}',
                    )
                elif "association_create" in url:
                    TestShowcase.association = "create"
                    result = json.dumps(datadict)
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_package_association_create"}'
                        % result,
                    )
                return mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_create(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "vocabulary" in url:
                    return vocabulary_mockshow(url, datadict)
                if url.endswith("show") or "list" in url:
                    return mockshow(url, datadict)
                if "create" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_create"}',
                    )

                resultdictcopy = copy.deepcopy(showcase_resultdict)
                resultdictcopy["state"] = datadict["state"]
                result = json.dumps(resultdictcopy)
                if datadict["title"] == "MyShowcase1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_create"}'
                        % result,
                    )
                if datadict["title"] == "MyShowcase2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_create"}',
                    )
                if datadict["title"] == "MyShowcase3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_create"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_create"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "vocabulary" in url:
                    return vocabulary_mockshow(url, datadict)
                if url.endswith("show") or "list" in url:
                    return mockshow(url, datadict)
                if "update" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not update", "__type": "TEST ERROR: Not Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_update"}',
                    )
                resultdictcopy = copy.deepcopy(showcase_resultdict)
                merge_two_dictionaries(resultdictcopy, datadict)

                result = json.dumps(resultdictcopy)
                if datadict["title"] == "MyShowcase1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_update"}'
                        % result,
                    )
                if datadict["title"] == "MyShowcase2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_update"}',
                    )
                if datadict["title"] == "MyShowcase3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_update"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_update"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_delete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if url.endswith("show") or "list" in url:
                    return mockshow(url, datadict)
                if "delete" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not delete", "__type": "TEST ERROR: Not Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_delete"}',
                    )
                if datadict["id"] == "05e392bf-04e0-4ca6-848c-4e87bba10746":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_delete"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=ckanext_showcase_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def allsearch(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mockallsearch(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    def test_read_from_hdx(self, configuration, read):
        showcase = Showcase.read_from_hdx(
            "05e392bf-04e0-4ca6-848c-4e87bba10746"
        )
        assert showcase["id"] == "05e392bf-04e0-4ca6-848c-4e87bba10746"
        assert showcase["title"] == "MyShowcase1"
        showcase = Showcase.read_from_hdx("TEST2")
        assert showcase is None
        showcase = Showcase.read_from_hdx("TEST3")
        assert showcase is None

    def test_create_in_hdx(self, configuration, post_create):
        showcase = Showcase()
        with pytest.raises(HDXError):
            showcase.create_in_hdx()
        showcase["id"] = "05e392bf-04e0-4ca6-848c-4e87bba10746"
        showcase["title"] = "LALA"
        with pytest.raises(HDXError):
            showcase.create_in_hdx()

        showcase_data = copy.deepcopy(TestShowcase.showcase_data)
        showcase = Showcase(showcase_data)
        showcase.create_in_hdx()
        assert showcase["id"] == "05e392bf-04e0-4ca6-848c-4e87bba10746"
        assert showcase["state"] == "active"

        showcase_data["title"] = "MyShowcase2"
        showcase = Showcase(showcase_data)
        with pytest.raises(HDXError):
            showcase.create_in_hdx()

        showcase_data["title"] = "MyShowcase3"
        showcase = Showcase(showcase_data)
        with pytest.raises(HDXError):
            showcase.create_in_hdx()

    def test_update_in_hdx(self, configuration, post_update):
        showcase = Showcase()
        showcase["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            showcase.update_in_hdx()
        showcase["title"] = "LALA"
        with pytest.raises(HDXError):
            showcase.update_in_hdx()

        showcase = Showcase.read_from_hdx(
            "05e392bf-04e0-4ca6-848c-4e87bba10746"
        )
        assert showcase["id"] == "05e392bf-04e0-4ca6-848c-4e87bba10746"
        assert showcase["title"] == "MyShowcase1"

        showcase["name"] = "TEST1"
        showcase["notes"] = "lalalala"
        showcase.update_in_hdx()
        assert showcase["name"] == "TEST1"
        assert showcase["notes"] == "lalalala"
        assert showcase["state"] == "active"
        expected = copy.deepcopy(showcase_resultdict)
        expected["notes"] = "lalalala"
        expected["name"] = "TEST1"
        expected["tags"] = [
            {
                "name": "economics",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            },
            {
                "name": "health",
                "vocabulary_id": "4381925f-0ae9-44a3-b30d-cae35598757b",
            },
        ]
        assert showcase.get_old_data_dict() == expected

        showcase["name"] = "NOTEXIST"
        with pytest.raises(HDXError):
            showcase.update_in_hdx()

        del showcase["name"]
        with pytest.raises(HDXError):
            showcase.update_in_hdx()

        showcase_data = copy.deepcopy(TestShowcase.showcase_data)
        showcase_data["title"] = "MyShowcase1"
        showcase_data["name"] = "TEST1"
        showcase = Showcase(showcase_data)
        showcase.create_in_hdx()
        assert showcase["name"] == "TEST1"
        assert showcase["notes"] == "My Showcase"
        assert showcase["state"] == "active"

    def test_delete_from_hdx(self, configuration, post_delete):
        showcase = Showcase.read_from_hdx(
            "05e392bf-04e0-4ca6-848c-4e87bba10746"
        )
        showcase.delete_from_hdx()
        del showcase["id"]
        with pytest.raises(HDXError):
            showcase.delete_from_hdx()

    def test_update_yaml(self, configuration, static_yaml):
        showcase_data = copy.deepcopy(TestShowcase.showcase_data)
        showcase = Showcase(showcase_data)
        assert showcase["title"] == "MyShowcase1"
        assert showcase["name"] == "showcase-1"
        showcase.update_from_yaml(static_yaml)
        assert showcase["title"] == "MyShowcase1"
        assert showcase["name"] == "my-showcase-1"

    def test_update_json(self, configuration, static_json):
        showcase_data = copy.deepcopy(TestShowcase.showcase_data)
        showcase = Showcase(showcase_data)
        assert showcase["title"] == "MyShowcase1"
        assert showcase["name"] == "showcase-1"
        showcase.update_from_json(static_json)
        assert showcase["title"] == "MyShowcase1"
        assert showcase["name"] == "new-showcase-1"

    def test_tags(self, configuration):
        showcase_data = copy.deepcopy(TestShowcase.showcase_data)
        showcase = Showcase(showcase_data)
        assert showcase.get_tags() == ["economy", "health"]
        showcase.add_tag("wash")
        assert showcase.get_tags() == [
            "economy",
            "health",
            "water sanitation and hygiene - wash",
        ]
        showcase.add_tags(["sanitation"])
        assert showcase.get_tags() == [
            "economy",
            "health",
            "water sanitation and hygiene - wash",
        ]
        result = showcase.remove_tag("water sanitation and hygiene - wash")
        assert result is True
        assert showcase.get_tags() == ["economy", "health"]
        showcase["tags"] = None
        result = showcase.remove_tag("wash")
        assert result is False

    def test_datasets(self, configuration, read):
        showcase = Showcase.read_from_hdx(
            "05e392bf-04e0-4ca6-848c-4e87bba10746"
        )
        datasets = showcase.get_datasets()
        assert len(datasets) == 10
        assert datasets[0].data == datasetsdict["results"][0]
        dict4 = copy.deepcopy(datasetsdict["results"][4])
        del dict4["resources"]
        assert datasets[4].data == dict4
        TestShowcase.association = None
        showcase.remove_dataset(datasets[0])
        assert TestShowcase.association == "delete"
        TestShowcase.association = None
        assert (
            showcase.add_dataset("a2f32edd-bac2-4940-aa58-49e565041055")
            is True
        )
        assert TestShowcase.association == "create"
        TestShowcase.association = None
        assert (
            showcase.add_datasets(
                [
                    {"id": "a2f32edd-bac2-4940-aa58-49e565041055"},
                    {"id": "6a5aebc1-f5a9-4842-8183-b8118228e71e"},
                ]
            )
            is False
        )
        assert TestShowcase.association == "create"
        TestShowcase.association = None
        assert showcase.add_dataset({"name": "TEST1"}) is True
        assert TestShowcase.association == "create"
        TestShowcase.association = None
        with pytest.raises(HDXError):
            showcase.add_dataset("123")
        with pytest.raises(HDXError):
            showcase.add_dataset(123)

    def test_search_in_hdx(self, configuration, allsearch):
        showcases = Showcase.search_in_hdx("ACLED")
        assert len(showcases) == 10
        showcases = Showcase.search_in_hdx("ACLED", offset=2, limit=6)
        assert len(showcases) == 6
        showcases = Showcase.search_in_hdx(
            fq="metadata_modified:[2018-01-01T00:00:00.000Z TO NOW]"
        )
        assert len(showcases) == 1
        showcases = Showcase.search_in_hdx("ajyhgr")
        assert len(showcases) == 0
        with pytest.raises(HDXError):
            Showcase.search_in_hdx('"')
        with pytest.raises(HDXError):
            Showcase.search_in_hdx("ACLED", rows=11)
        with pytest.raises(HDXError):
            # Test returned row counts per page mismatch (wrong count of 6 purposely in mocksearch)
            Showcase.search_in_hdx("ACLED", page_size=5)

    def test_get_all_showcases(self, configuration, allsearch):
        showcases = Showcase.get_all_showcases()
        assert len(showcases) == 20



"""Organization Tests"""
import copy
import json
from os.path import join

import pytest
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.loader import load_yaml

from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXError
from hdx.data.organization import Organization
from hdx.data.user import User

from . import MockResponse, organization_data, user_data
from .test_user import user_mockshow

resultdict = load_yaml(
    join("tests", "fixtures", "organization_show_results.yml")
)

organization_list = [
    "acaps",
    "accion-contra-el-hambre-acf-spain",
    "acf-west-africa",
    "acled",
    "acted",
    "action-contre-la-faim",
    "adeso",
    "afd",
    "afdb",
    "afghanistan-protection-cluster",
]
searchdict = load_yaml(join("tests", "fixtures", "dataset_search_results.yml"))

organization_autocomplete = [
    {
        "title": "INNAGO",
        "id": "5a63012e-6c41-420c-8c33-e84b277fdc90",
        "name": "innago",
    }
]


def organization_mockshow(url, datadict):
    if "show" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not show", "__type": "TEST ERROR: Not Show Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_show"}',
        )
    result = json.dumps(resultdict)
    if (
        datadict["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        or datadict["id"] == "TEST1"
    ):
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_show"}'
            % result,
        )
    if datadict["id"] == "TEST2":
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_show"}',
        )
    if datadict["id"] == "TEST3":
        return MockResponse(
            200,
            '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_show"}',
        )
    return MockResponse(
        404,
        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_show"}',
    )


def mocklist(url):
    if "list" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not all", "__type": "TEST ERROR: Not All Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_list"}',
        )
    return MockResponse(
        200,
        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_list"}'
        % json.dumps(organization_list),
    )


def mockgetdatasets(url, datadict):
    if "search" not in url:
        return MockResponse(
            404,
            '{"success": false, "error": {"message": "TEST ERROR: Not search", "__type": "TEST ERROR: Not Search Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}',
        )
    if datadict["fq"] == "organization:acled":
        return MockResponse(
            200,
            '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=package_search"}'
            % json.dumps(searchdict),
        )


class TestOrganization:
    @pytest.fixture(scope="class")
    def static_yaml(self):
        return join(
            "tests", "fixtures", "config", "hdx_organization_static.yml"
        )

    @pytest.fixture(scope="class")
    def static_json(self):
        return join(
            "tests", "fixtures", "config", "hdx_organization_static.json"
        )

    @pytest.fixture(scope="function")
    def read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return organization_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_create(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return organization_mockshow(url, datadict)
                if "create" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not create", "__type": "TEST ERROR: Not Create Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_create"}',
                    )

                resultdictcopy = copy.deepcopy(resultdict)
                resultdictcopy["state"] = datadict["state"]
                result = json.dumps(resultdictcopy)
                if datadict["name"] == "MyOrganization1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_create"}'
                        % result,
                    )
                if datadict["name"] == "MyOrganization2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_create"}',
                    )
                if datadict["name"] == "MyOrganization3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_create"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_create"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_update(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                if "show" in url:
                    return organization_mockshow(url, datadict)
                if "update" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not update", "__type": "TEST ERROR: Not Update Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_update"}',
                    )
                resultdictcopy = copy.deepcopy(resultdict)
                merge_two_dictionaries(resultdictcopy, datadict)

                result = json.dumps(resultdictcopy)
                if datadict["name"] == "MyOrganization1":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_update"}'
                        % result,
                    )
                if datadict["name"] == "MyOrganization2":
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_update"}',
                    )
                if datadict["name"] == "MyOrganization3":
                    return MockResponse(
                        200,
                        '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_update"}',
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_update"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_delete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "show" in url:
                    return organization_mockshow(url, datadict)
                if "delete" not in url:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not delete", "__type": "TEST ERROR: Not Delete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_delete"}',
                    )
                if datadict["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19":
                    return MockResponse(
                        200,
                        '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_delete"}'
                        % decodedata,
                    )

                return MockResponse(
                    404,
                    '{"success": false, "error": {"message": "Not found", "__type": "Not Found Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_delete"}',
                )

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_list(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                json.loads(data.decode("utf-8"))
                return mocklist(url)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def user_read(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return user_mockshow(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def datasets_get(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                datadict = json.loads(data.decode("utf-8"))
                return mockgetdatasets(url, datadict)

        Configuration.read().remoteckan().session = MockSession()

    @pytest.fixture(scope="function")
    def post_autocomplete(self):
        class MockSession:
            @staticmethod
            def post(url, data, headers, files, allow_redirects, auth=None):
                decodedata = data.decode("utf-8")
                datadict = json.loads(decodedata)
                if "autocomplete" not in url or "innago" not in datadict["q"]:
                    return MockResponse(
                        404,
                        '{"success": false, "error": {"message": "TEST ERROR: Not autocomplete", "__type": "TEST ERROR: Not Autocomplete Error"}, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_autocomplete"}',
                    )
                result = json.dumps(organization_autocomplete)
                return MockResponse(
                    200,
                    '{"success": true, "result": %s, "help": "http://test-data.humdata.org/api/3/action/help_show?name=organization_autocomplete"}'
                    % result,
                )

        Configuration.read().remoteckan().session = MockSession()

    def test_read_from_hdx(self, configuration, read, mocksmtp):
        organization = Organization.read_from_hdx(
            "b67e6c74-c185-4f43-b561-0e114a736f19"
        )
        assert organization["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        assert organization["name"] == "acled"
        organization = Organization.read_from_hdx("TEST2")
        assert organization is None
        organization = Organization.read_from_hdx("TEST3")
        assert organization is None

    def test_create_in_hdx(self, configuration, post_create):
        organization = Organization()
        with pytest.raises(HDXError):
            organization.create_in_hdx()
        organization["id"] = "b67e6c74-c185-4f43-b561-0e114a736f19"
        organization["name"] = "LALA"
        with pytest.raises(HDXError):
            organization.create_in_hdx()

        org_data = copy.deepcopy(organization_data)
        organization = Organization(org_data)
        organization.create_in_hdx()
        assert organization["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        assert organization["state"] == "active"

        org_data["name"] = "MyOrganization2"
        organization = Organization(org_data)
        with pytest.raises(HDXError):
            organization.create_in_hdx()

        org_data["name"] = "MyOrganization3"
        organization = Organization(org_data)
        with pytest.raises(HDXError):
            organization.create_in_hdx()

    def test_update_in_hdx(self, configuration, post_update):
        organization = Organization()
        organization["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            organization.update_in_hdx()
        organization["name"] = "LALA"
        with pytest.raises(HDXError):
            organization.update_in_hdx()

        organization = Organization.read_from_hdx(
            "b67e6c74-c185-4f43-b561-0e114a736f19"
        )
        assert organization["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        assert organization["name"] == "acled"

        organization["description"] = "Humanitarian work"
        organization["id"] = "b67e6c74-c185-4f43-b561-0e114a736f19"
        organization["name"] = "MyOrganization1"
        organization.update_in_hdx()
        assert organization["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        assert organization["description"] == "Humanitarian work"
        assert organization["state"] == "active"

        organization["id"] = "NOTEXIST"
        with pytest.raises(HDXError):
            organization.update_in_hdx()

        del organization["id"]
        with pytest.raises(HDXError):
            organization.update_in_hdx()

        org_data = copy.deepcopy(organization_data)
        org_data["name"] = "MyOrganization1"
        org_data["id"] = "b67e6c74-c185-4f43-b561-0e114a736f19"
        organization = Organization(org_data)
        organization.create_in_hdx()
        assert organization["id"] == "b67e6c74-c185-4f43-b561-0e114a736f19"
        assert organization["name"] == "MyOrganization1"
        assert organization["state"] == "active"

    def test_delete_from_hdx(self, configuration, post_delete):
        organization = Organization.read_from_hdx(
            "b67e6c74-c185-4f43-b561-0e114a736f19"
        )
        organization.delete_from_hdx()
        del organization["id"]
        with pytest.raises(HDXError):
            organization.delete_from_hdx()

    def test_update_yaml(self, configuration, static_yaml):
        org_data = copy.deepcopy(organization_data)
        organization = Organization(org_data)
        assert organization["name"] == "MyOrganization1"
        assert organization["description"] == "We do humanitarian work"
        organization.update_from_yaml(static_yaml)
        assert organization["name"] == "MyOrg1"
        assert organization["description"] == "Humanitarian organisation"

    def test_update_json(self, configuration, static_json):
        org_data = copy.deepcopy(organization_data)
        organization = Organization(org_data)
        assert organization["name"] == "MyOrganization1"
        assert organization["description"] == "We do humanitarian work"
        organization.update_from_json(static_json)
        assert organization["name"] == "MyOrg1"
        assert organization["description"] == "other"

    def test_get_all_organizations(self, configuration, post_list):
        organizations = Organization.get_all_organization_names()
        assert len(organizations) == 10

    def test_users(self, configuration, user_read):
        org_data = copy.deepcopy(resultdict)
        organization = Organization(org_data)
        organization["users"][0]["name"] = "MyUser1"
        users = organization.get_users()
        assert len(users) == 1
        assert users[0]["name"] == "MyUser1"
        organization.remove_user("9f3e9973-7dbe-4c65-8820-f48578e3ffea")
        users = organization.get_users()
        assert len(users) == 0
        userdata_copy = copy.deepcopy(user_data)
        userdata_copy["id"] = "9f3e9973-7dbe-4c65-8820-f48578e3ffea"
        user = User(userdata_copy)
        user["name"] = "TEST1"
        user["capacity"] = "member"
        del organization["users"]
        organization.add_update_users([user])
        users = organization.get_users("member")
        assert len(users) == 1
        organization.remove_user(users[0])
        users = organization.get_users()
        assert len(users) == 0
        organization.add_update_users([user])
        users = organization.get_users("member")
        assert len(users) == 1
        organization.add_update_users(["MyUser1"], "member")
        users = organization.get_users("member")
        assert len(users) == 2
        assert users[0]["name"] == "MyUser1"
        organization.remove_user(users[0].data)
        users = organization.get_users("member")
        assert len(users) == 1
        organization.add_update_users([user], "editor")
        organization.remove_user("")
        users = organization.get_users("editor")
        assert len(users) == 1
        organization.remove_user("NOT_EXIST")
        users = organization.get_users("editor")
        assert len(users) == 1
        users = organization.get_users("admin")
        assert len(users) == 0
        del organization["users"][0]["id"]
        users = organization.get_users("member")
        assert len(users) == 1
        organization.add_update_user({"name": "TEST1"}, "member")
        users = organization.get_users("member")
        assert len(users) == 2
        with pytest.raises(HDXError):
            organization.add_update_users(123)
        with pytest.raises(HDXError):
            organization.add_update_user(123)
        with pytest.raises(HDXError):
            organization.remove_user(123)

    def test_get_datasets(self, configuration, datasets_get):
        org_data = copy.deepcopy(resultdict)
        organization = Organization(org_data)
        datasets = organization.get_datasets()
        assert len(datasets) == 10

    def test_autocomplete(self, configuration, post_autocomplete):
        assert Organization.autocomplete("innago") == organization_autocomplete



"""Simple Facade Tests"""
import sys

import pytest
from hdx.utilities.useragent import UserAgent

from hdx.api.configuration import ConfigurationError
from hdx.facades.infer_arguments import facade

from . import my_testfnia, testresult


class TestInferArguments:
    def test_facade(
        self,
        monkeypatch,
        hdx_config_yaml,
        hdx_config_json,
        project_config_yaml,
    ):
        UserAgent.clear_global()
        mydata = "hello"
        monkeypatch.setattr(
            sys,
            "argv",
            [
                "test",
                "--mydata",
                mydata,
                "--hdx-site",
                "prod",
                "--hdx-key",
                "123",
                "--user-agent",
                "test",
            ],
        )
        testresult.actual_result = None
        facade(my_testfnia)
        assert testresult.actual_result == mydata

        UserAgent.clear_global()
        mydata = "hello"
        monkeypatch.setattr(
            sys,
            "argv",
            [
                "test",
                "--mydata",
                mydata,
                "--hdx-site",
                "prod",
                "--user-agent",
                "test",
            ],
        )
        testresult.actual_result = None
        facade(my_testfnia, hdx_site="stage", hdx_key="123")
        assert testresult.actual_result == mydata

        UserAgent.clear_global()
        monkeypatch.setattr(
            sys,
            "argv",
            [
                "test",
                "--mydata",
                mydata,
                "--hdx-site",
                "prod",
                "--user-agent",
                "test",
            ],
        )
        with pytest.raises(FileNotFoundError):
            facade(my_testfnia, hdx_config_yaml="NOT EXIST")

        UserAgent.clear_global()
        monkeypatch.setattr(
            sys,
            "argv",
            [
                "test",
                "--mydata",
                mydata,
                "--hdx-site",
                "prod",
                "--user-agent",
                "test",
            ],
        )
        with pytest.raises(ConfigurationError):
            facade(
                my_testfnia,
                hdx_config_yaml=hdx_config_yaml,
                hdx_config_json=hdx_config_json,
            )

        UserAgent.clear_global()
        monkeypatch.setattr(
            sys,
            "argv",
            [
                "test",
                "--mydata",
                mydata,
                "--hdx-site",
                "prod",
                "--hdx-key",
                "123",
                "--user-agent",
                "test",
                "--lala",
                "what",
            ],
        )
        with pytest.raises(SystemExit):
            facade(my_testfnia)
        UserAgent.clear_global()



"""Showcase class containing all logic for creating, checking, and updating showcases."""
import logging
import sys
from os.path import join
from typing import Any, Dict, List, Optional, Tuple, Union

from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.uuid import is_valid_uuid

import hdx.data.dataset
import hdx.data.vocabulary
from hdx.api.configuration import Configuration
from hdx.data.hdxobject import HDXObject

logger = logging.getLogger(__name__)


class Showcase(HDXObject):
    """Showcase class containing all logic for creating, checking, and updating showcases.

    Args:
        initial_data (Optional[Dict]): Initial showcase metadata dictionary. Defaults to None.
        configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
    """

    max_int = sys.maxsize
    dataset_ids_field = "dataset_ids"

    def __init__(
        self,
        initial_data: Optional[Dict] = None,
        configuration: Optional[Configuration] = None,
    ) -> None:
        if not initial_data:
            initial_data = dict()
        super().__init__(initial_data, configuration=configuration)

    @staticmethod
    def actions() -> Dict[str, str]:
        """Dictionary of actions that can be performed on object

        Returns:
            Dict[str, str]: Dictionary of actions that can be performed on object
        """
        return {
            "show": "ckanext_showcase_show",
            "update": "ckanext_showcase_update",
            "create": "ckanext_showcase_create",
            "delete": "ckanext_showcase_delete",
            "list": "ckanext_showcase_list",
            "associate": "ckanext_showcase_package_association_create",
            "disassociate": "ckanext_showcase_package_association_delete",
            "list_datasets": "ckanext_showcase_package_list",
            "list_showcases": "ckanext_package_showcase_list",
        }

    def update_from_yaml(
        self, path: str = join("config", "hdx_showcase_static.yml")
    ) -> None:
        """Update showcase metadata with static metadata from YAML file

        Args:
            path (Optional[str]): Path to YAML dataset metadata. Defaults to config/hdx_showcase_static.yml.

        Returns:
            None
        """
        super().update_from_yaml(path)

    def update_from_json(
        self, path: str = join("config", "hdx_showcase_static.json")
    ) -> None:
        """Update showcase metadata with static metadata from JSON file

        Args:
            path (Optional[str]): Path to JSON dataset metadata. Defaults to config/hdx_showcase_static.json.

        Returns:
            None
        """
        super().update_from_json(path)

    @classmethod
    def read_from_hdx(
        cls, identifier: str, configuration: Optional[Configuration] = None
    ) -> Optional["Showcase"]:
        """Reads the showcase given by identifier from HDX and returns Showcase object

        Args:
            identifier (str): Identifier of showcase
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.

        Returns:
            Optional[Showcase]: Showcase object if successful read, None if not
        """
        return cls._read_from_hdx_class("showcase", identifier, configuration)

    def check_required_fields(self, ignore_fields: List[str] = list()) -> None:
        """Check that metadata for showcase is complete. The parameter ignore_fields should
        be set if required to any fields that should be ignored for the particular operation.

        Args:
            ignore_fields (List[str]): Fields to ignore. Default is [].

        Returns:
            None
        """
        self._check_required_fields("showcase", ignore_fields)

    def update_in_hdx(self, **kwargs: Any) -> None:
        """Check if showcase exists in HDX and if so, update it

        Returns:
            None
        """
        self._check_load_existing_object("showcase", "name")
        # We load an existing object even though it may well have been loaded already
        # to prevent an admittedly unlikely race condition where someone has updated
        # the object in the intervening time
        merge_two_dictionaries(self.data, self.old_data)
        self.clean_tags()
        self._hdx_update("showcase", "name", force_active=True, **kwargs)
        self._update_in_hdx("showcase", "name", **kwargs)

    def create_in_hdx(self, **kwargs: Any) -> None:
        """Check if showcase exists in HDX and if so, update it, otherwise create it

        Returns:
            None
        """
        if "ignore_check" not in kwargs:  # allow ignoring of field checks
            self.check_required_fields()
        if "name" in self.data and self._load_from_hdx(
            "showcase", self.data["name"]
        ):
            logger.warning(
                f"{'showcase'} exists. Updating {self.data['name']}"
            )
            merge_two_dictionaries(self.data, self.old_data)
            self.clean_tags()
            self._hdx_update("showcase", "name", force_active=True, **kwargs)
        else:
            self.clean_tags()
            self._save_to_hdx("create", "title", force_active=True)

        self._create_in_hdx("showcase", "name", "title", **kwargs)

    def delete_from_hdx(self) -> None:
        """Deletes a showcase from HDX.

        Returns:
            None
        """
        self._delete_from_hdx("showcase", "id")

    def get_tags(self) -> List[str]:
        """Return the dataset's list of tags

        Returns:
            List[str]: List of tags or [] if there are none
        """
        return self._get_tags()

    def add_tag(
        self, tag: str, log_deleted: bool = True
    ) -> Tuple[List[str], List[str]]:
        """Add a tag

        Args:
            tag (str): Tag to add
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of added tags and list of deleted tags and tags not added
        """
        return hdx.data.vocabulary.Vocabulary.add_mapped_tag(
            self, tag, log_deleted=log_deleted
        )

    def add_tags(
        self, tags: List[str], log_deleted: bool = True
    ) -> Tuple[List[str], List[str]]:
        """Add a list of tags

        Args:
            tags (List[str]): List of tags to add
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of added tags and list of deleted tags and tags not added
        """
        return hdx.data.vocabulary.Vocabulary.add_mapped_tags(
            self, tags, log_deleted=log_deleted
        )

    def clean_tags(
        self, log_deleted: bool = True
    ) -> Tuple[List[str], List[str]]:
        """Clean tags in an HDX object according to tags cleanup spreadsheet

        Args:
            log_deleted (bool): Whether to log informational messages about deleted tags. Defaults to True.

        Returns:
            Tuple[List[str], List[str]]: Tuple containing list of mapped tags and list of deleted tags and tags not added
        """
        return hdx.data.vocabulary.Vocabulary.clean_tags(
            self, log_deleted=log_deleted
        )

    def remove_tag(self, tag: str) -> bool:
        """Remove a tag

        Args:
            tag (str): Tag to remove

        Returns:
            bool: True if tag removed or False if not
        """
        return self._remove_hdxobject(
            self.data.get("tags"), tag.lower(), matchon="name"
        )

    def get_datasets(self) -> List["Dataset"]:  # noqa: F821
        """Get any datasets in the showcase

        Returns:
            List[Dataset]: List of datasets
        """
        assoc_result, datasets_dicts = self._read_from_hdx(
            "showcase",
            self.data["id"],
            fieldname="showcase_id",
            action=self.actions()["list_datasets"],
        )
        datasets = list()
        if assoc_result:
            for dataset_dict in datasets_dicts:
                dataset = hdx.data.dataset.Dataset(
                    dataset_dict, configuration=self.configuration
                )
                datasets.append(dataset)
        return datasets

    def _get_showcase_dataset_dict(
        self, dataset: Union["Dataset", Dict, str]  # noqa: F821
    ) -> Dict:
        """Get showcase dataset dict

        Args:
            showcase (Union[Showcase,Dict,str]): Either a showcase id or Showcase metadata from a Showcase object or dictionary

        Returns:
            Dict: showcase dataset dict
        """
        if isinstance(dataset, hdx.data.dataset.Dataset) or isinstance(
            dataset, dict
        ):
            if "id" not in dataset:
                dataset = hdx.data.dataset.Dataset.read_from_hdx(
                    dataset["name"]
                )
            dataset = dataset["id"]
        elif not isinstance(dataset, str):
            raise hdx.data.hdxobject.HDXError(
                f"Type {type(dataset).__name__} cannot be added as a dataset!"
            )
        if is_valid_uuid(dataset) is False:
            raise hdx.data.hdxobject.HDXError(
                f"{dataset} is not a valid dataset id!"
            )
        return {"showcase_id": self.data["id"], "package_id": dataset}

    def add_dataset(
        self,
        dataset: Union["Dataset", Dict, str],  # noqa: F821
        datasets_to_check: List["Dataset"] = None,  # noqa: F821
    ) -> bool:
        """Add a dataset

        Args:
            dataset (Union[Dataset,Dict,str]): Either a dataset id or dataset metadata either from a Dataset object or a dictionary
            datasets_to_check (List[Dataset]): List of datasets against which to check existence of dataset. Defaults to datasets in showcase.

        Returns:
            bool: True if the dataset was added, False if already present
        """
        showcase_dataset = self._get_showcase_dataset_dict(dataset)
        if datasets_to_check is None:
            datasets_to_check = self.get_datasets()
        for dataset in datasets_to_check:
            if showcase_dataset["package_id"] == dataset["id"]:
                return False
        self._write_to_hdx("associate", showcase_dataset, "package_id")
        return True

    def add_datasets(
        self,
        datasets: List[Union["Dataset", Dict, str]],  # noqa: F821
        datasets_to_check: List["Dataset"] = None,  # noqa: F821
    ) -> bool:
        """Add multiple datasets

        Args:
            datasets (List[Union[Dataset,Dict,str]]): A list of either dataset ids or dataset metadata from Dataset objects or dictionaries
            datasets_to_check (List[Dataset]): List of datasets against which to check existence of dataset. Defaults to datasets in showcase.

        Returns:
            bool: True if all datasets added or False if any already present
        """
        if datasets_to_check is None:
            datasets_to_check = self.get_datasets()
        alldatasetsadded = True
        for dataset in datasets:
            if not self.add_dataset(
                dataset, datasets_to_check=datasets_to_check
            ):
                alldatasetsadded = False
        return alldatasetsadded

    def remove_dataset(
        self, dataset: Union["Dataset", Dict, str]  # noqa: F821
    ) -> None:
        """Remove a dataset

        Args:
            dataset (Union[Dataset,Dict,str]): Either a dataset id or dataset metadata either from a Dataset object or a dictionary

        Returns:
            None
        """
        self._write_to_hdx(
            "disassociate",
            self._get_showcase_dataset_dict(dataset),
            "package_id",
        )

    @classmethod
    def search_in_hdx(
        cls,
        query: Optional[str] = "*:*",
        configuration: Optional[Configuration] = None,
        page_size: int = 1000,
        **kwargs: Any,
    ) -> List["Showcase"]:
        """Searches for datasets in HDX

        Args:
            query (Optional[str]): Query (in Solr format). Defaults to '*:*'.
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
            page_size (int): Size of page to return. Defaults to 1000.
            **kwargs: See below
            fq (string): Any filter queries to apply
            sort (string): Sorting of the search results. Defaults to 'relevance asc, metadata_modified desc'.
            rows (int): Number of matching rows to return. Defaults to all datasets (sys.maxsize).
            start (int): Offset in the complete result for where the set of returned datasets should begin
            facet (string): Whether to enable faceted results. Default to True.
            facet.mincount (int): Minimum counts for facet fields should be included in the results
            facet.limit (int): Maximum number of values the facet fields return (- = unlimited). Defaults to 50.
            facet.field (List[str]): Fields to facet upon. Default is empty.
            use_default_schema (bool): Use default package schema instead of custom schema. Defaults to False.

        Returns:
            List[Dataset]: list of datasets resulting from query
        """
        curfq = kwargs.get("fq")
        kwargs["fq"] = "dataset_type:showcase"
        if curfq:
            kwargs["fq"] = f"{kwargs['fq']} AND {curfq}"
        datasets = hdx.data.dataset.Dataset.search_in_hdx(
            query=query,
            configuration=configuration,
            page_size=page_size,
            **kwargs,
        )
        showcases = list()
        for dataset in datasets:
            showcase = Showcase(configuration=configuration)
            showcase.data = dataset.data
            showcase.old_data = dataset.old_data
            showcases.append(showcase)
        return showcases

    @classmethod
    def get_all_showcases(
        cls,
        configuration: Optional[Configuration] = None,
        page_size: int = 1000,
        **kwargs: Any,
    ) -> List["Showcase"]:
        """Get all showcases in HDX

        Args:
            configuration (Optional[Configuration]): HDX configuration. Defaults to global configuration.
            page_size (int): Size of page to return. Defaults to 1000.
            **kwargs: See below
            rows (int): Number of rows to return. Defaults to all showcases (sys.maxsize)
            start (int): Offset in the complete result for where the set of returned showcases should begin

        Returns:
            List[Showcase]: list of all showcases in HDX
        """
        return cls.search_in_hdx(
            configuration=configuration, page_size=page_size, **kwargs
        )



"""Dataset Title Helper Tests"""
from datetime import datetime, timezone

import pytest

from hdx.data.dataset_title_helper import DatasetTitleHelper


class TestDatasetTitleHelper:
    @pytest.fixture
    def expected_ranges_2019(self):
        return [
            (
                datetime(2019, 7, 1, 0, 0, tzinfo=timezone.utc),
                datetime(2019, 7, 31, 23, 59, 59, tzinfo=timezone.utc),
            )
        ]

    def test_fuzzy_match_dates_in_title(self, expected_ranges_2019):
        ignore_wrong_years = list()
        ranges = list()
        assert (
            DatasetTitleHelper.fuzzy_match_dates_in_title(
                "Myanmar Town July 2019", ranges, ignore_wrong_years
            )
            == "Myanmar Town"
        )
        assert ranges == expected_ranges_2019
        ranges = list()
        assert (
            DatasetTitleHelper.fuzzy_match_dates_in_title(
                "Myanmar Town 2019 July", ranges, ignore_wrong_years
            )
            == "Myanmar Town"
        )
        assert ranges == expected_ranges_2019

    def test_get_date_from_title(self, expected_ranges_2019):
        assert DatasetTitleHelper.get_dates_from_title(
            "Myanmar Self Administered Regions Boundaries MIMU v9.2.1"
        ) == (
            "Myanmar Self Administered Regions Boundaries MIMU v9.2.1",
            list(),
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Myanmar Town 2019 July"
        ) == (
            "Myanmar Town",
            expected_ranges_2019,
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Formal Sector School Location Upper Myanmar (  2019   )"
        ) == (
            "Formal Sector School Location Upper Myanmar",
            [
                (
                    datetime(2019, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "ICA Armenia, 2017 - Drought Risk, 1981-2015"
        ) == (
            "ICA Armenia - Drought Risk",
            [
                (
                    datetime(1981, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2015, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2017, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2017, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "ICA Sudan, 2018 - Land Degradation, 2001-2013"
        ) == (
            "ICA Sudan - Land Degradation",
            [
                (
                    datetime(2001, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2013, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2018, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2018, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Central African Republic, Bridges, January 2019"
        ) == (
            "Central African Republic, Bridges",
            [
                (
                    datetime(2019, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 1, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Afghanistan:District Accessibility for WFP and Partners Staff as of 05 May 2019"
        ) == (
            "Afghanistan:District Accessibility for WFP and Partners Staff",
            [
                (
                    datetime(2019, 5, 5, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 5, 5, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Tanintharyi Region Land Cover - March 2016 (Original)"
        ) == (
            "Tanintharyi Region Land Cover (Original)",
            [
                (
                    datetime(2016, 3, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2016, 3, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Kachin State and Sagaing Region 2002-2014 Forest Cover Change"
        ) == (
            "Kachin State and Sagaing Region Forest Cover Change",
            [
                (
                    datetime(2002, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2014, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Ward boundaries Yangon City_mimu_v8_1"
        ) == ("Ward boundaries Yangon City_mimu_v8_1", list())
        assert DatasetTitleHelper.get_dates_from_title(
            "Mon_State_Village_Tract_Boundaries"
        ) == ("Mon_State_Village_Tract_Boundaries", list())
        assert DatasetTitleHelper.get_dates_from_title(
            "ICA Afghanistan, 2019 - Landslide hazard, 2013"
        ) == (
            "ICA Afghanistan - Landslide hazard",
            [
                (
                    datetime(2013, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2013, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2019, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Afghanistan Percentage of Food Insecure Population Based on Combined Food Consumption Score and Coping Strategy Index by Province - ALCS 2013/14"
        ) == (
            "Afghanistan Percentage of Food Insecure Population Based on Combined Food Consumption Score and Coping Strategy Index by Province - ALCS",
            [
                (
                    datetime(2013, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2014, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title("ALCS 2000/10") == (
            "ALCS",
            [
                (
                    datetime(2000, 10, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2000, 10, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "ALCS 2014/13"
        ) == (  # not a month and range going down
            "ALCS 2014/13",
            list(),
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Mon_State_Village_Tract_Boundaries 9999"
        ) == ("Mon_State_Village_Tract_Boundaries 9999", list())
        assert DatasetTitleHelper.get_dates_from_title(
            "Mon_State_Village_Tract_Boundaries 10/12/01 lala"
        ) == ("Mon_State_Village_Tract_Boundaries 10/12/01 lala", list())
        # It's the "Mon" that makes an extra date component that causes it to ignore the date (correctly)
        assert DatasetTitleHelper.get_dates_from_title(
            "State_Village_Tract_Boundaries 10/12/01 lala"
        ) == (
            "State_Village_Tract_Boundaries lala",
            [
                (
                    datetime(2001, 12, 10, 0, 0, tzinfo=timezone.utc),
                    datetime(2001, 12, 10, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Crops production (2016) - Tajikistan Vulnerability & Resilience Atlas, 2019"
        ) == (
            "Crops production - Tajikistan Vulnerability & Resilience Atlas",
            [
                (
                    datetime(2016, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2016, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2019, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Location of partners as of Feb. 5, 2019"
        ) == (
            "Location of partners",
            [
                (
                    datetime(2019, 2, 5, 0, 0, tzinfo=timezone.utc),
                    datetime(2019, 2, 5, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "ICA Armenia, 2016 & 2017 - Land Degradation, 2001-2012"
        ) == (
            "ICA Armenia - Land Degradation",
            [
                (
                    datetime(2001, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2012, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2016, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2017, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "ICA Afghanistan, 2016 - Food Insecurity Risk, 2007/08-2014"
        ) == (
            "ICA Afghanistan - Food Insecurity Risk",
            [
                (
                    datetime(2007, 8, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2014, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2016, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2016, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Risk, 2020/19-2014"
        ) == (
            "Risk, 2020/19",
            [
                (
                    datetime(2014, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2014, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "south sudan access constraints shp for 20200124"
        ) == (
            "south sudan access constraints shp",
            [
                (
                    datetime(2020, 1, 24, 0, 0, tzinfo=timezone.utc),
                    datetime(2020, 1, 24, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Cambodia Flood Extent in 2011"
        ) == (
            "Cambodia Flood Extent",
            [
                (
                    datetime(2011, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2011, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                )
            ],
        )
        assert DatasetTitleHelper.get_dates_from_title(
            "Access: Proportion of the population consuming less than 2100 kcal per day (average of 2011-2013), National Statistics Committee 2013"
        ) == (
            "Access: Proportion of the population consuming less than 2100 kcal per day (average), National Statistics Committee",
            [
                (
                    datetime(2011, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2013, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
                (
                    datetime(2013, 1, 1, 0, 0, tzinfo=timezone.utc),
                    datetime(2013, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
                ),
            ],
        )



"""Simple Facade Tests"""
import pytest
from hdx.utilities.useragent import UserAgent, UserAgentError

from hdx.api import __version__
from hdx.facades.keyword_arguments import facade

from . import my_testfnkw, testresult


class TestKeywordArguments:
    def test_facade(self, monkeypatch, hdx_config_yaml, project_config_yaml):
        UserAgent.clear_global()
        my_user_agent = "test"
        testresult.actual_result = None
        facade(
            my_testfnkw,
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="site",
        )
        assert testresult.actual_result == "https://data.humdata.org"
        UserAgent.clear_global()
        testresult.actual_result = None
        facade(
            my_testfnkw,
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="agent",
        )
        assert (
            testresult.actual_result
            == f"HDXPythonLibrary/{__version__}-{my_user_agent}"
        )
        UserAgent.clear_global()
        testresult.actual_result = None
        my_user_agent = "lala"
        monkeypatch.setenv("USER_AGENT", my_user_agent)
        facade(
            my_testfnkw,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="agent",
        )
        assert (
            testresult.actual_result
            == f"HDXPythonLibrary/{__version__}-{my_user_agent}"
        )
        UserAgent.clear_global()
        testresult.actual_result = None
        facade(
            my_testfnkw,
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="agent",
        )
        assert (
            testresult.actual_result
            == f"HDXPythonLibrary/{__version__}-{my_user_agent}"
        )
        UserAgent.clear_global()
        testresult.actual_result = None
        my_preprefix = "haha"
        monkeypatch.setenv("PREPREFIX", my_preprefix)
        facade(
            my_testfnkw,
            user_agent="test",
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="agent",
        )
        assert (
            testresult.actual_result
            == f"{my_preprefix}:HDXPythonLibrary/{__version__}-{my_user_agent}"
        )
        UserAgent.clear_global()
        testresult.actual_result = None
        my_test_key = "1234"
        facade(
            my_testfnkw,
            hdx_key=my_test_key,
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="api",
        )
        assert testresult.actual_result == my_test_key
        UserAgent.clear_global()
        testresult.actual_result = None
        monkeypatch.setenv("HDX_KEY", my_test_key)
        facade(
            my_testfnkw,
            hdx_key="aaaa",
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="api",
        )
        assert testresult.actual_result == my_test_key
        UserAgent.clear_global()
        testresult.actual_result = None
        my_test_hdxsite = "stage"
        facade(
            my_testfnkw,
            hdx_site=my_test_hdxsite,
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="site",
        )
        assert (
            testresult.actual_result
            == f"https://{my_test_hdxsite}.data-humdata-org.ahconu.org"
        )
        UserAgent.clear_global()
        testresult.actual_result = None
        monkeypatch.setenv("HDX_SITE", my_test_hdxsite)
        facade(
            my_testfnkw,
            hdx_site="feature",
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="site",
        )
        assert (
            testresult.actual_result
            == f"https://{my_test_hdxsite}.data-humdata-org.ahconu.org"
        )
        UserAgent.clear_global()
        my_test_hdxurl = "http://other-data.humdata.org"
        monkeypatch.setenv("HDX_URL", my_test_hdxurl)
        facade(
            my_testfnkw,
            hdx_site="feature",
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="site",
        )
        assert testresult.actual_result == my_test_hdxurl
        UserAgent.clear_global()
        my_test_hdxurl2 = "http://other-data.humdata.org/"
        monkeypatch.setenv("HDX_URL", my_test_hdxurl2)
        facade(
            my_testfnkw,
            hdx_site="feature",
            user_agent=my_user_agent,
            hdx_config_yaml=hdx_config_yaml,
            project_config_yaml=project_config_yaml,
            fn="site",
        )
        assert testresult.actual_result == my_test_hdxurl
        UserAgent.clear_global()

    def test_exception(self, hdx_config_yaml, project_config_yaml):
        UserAgent.clear_global()
        testresult.actual_result = None
        with pytest.raises(ValueError):
            facade(
                my_testfnkw,
                user_agent="test",
                hdx_config_yaml=hdx_config_yaml,
                project_config_yaml=project_config_yaml,
                fn="exc",
            )
        UserAgent.clear_global()
        with pytest.raises(UserAgentError):
            facade(
                my_testfnkw,
                hdx_config_yaml=hdx_config_yaml,
                project_config_yaml=project_config_yaml,
                fn="agent",
            )
        UserAgent.clear_global()



"""Configuration for HDX"""
import logging
import os
from base64 import b64decode
from collections import UserDict
from os.path import expanduser, isfile, join
from typing import Any, Dict, Optional, Tuple

import ckanapi
import requests
from hdx.utilities.dictandlist import merge_two_dictionaries
from hdx.utilities.email import Email
from hdx.utilities.loader import load_json, load_yaml
from hdx.utilities.path import script_dir_plus_file
from hdx.utilities.session import get_session
from hdx.utilities.useragent import UserAgent, UserAgentError

from . import __version__

logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


class Configuration(UserDict):
    """Configuration for HDX

    Args:
        **kwargs: See below
        user_agent (str): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
        user_agent_config_yaml (str): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
        user_agent_lookup (str): Lookup key for YAML. Ignored if user_agent supplied.
        hdx_url (str): HDX url to use. Overrides hdx_site.
        hdx_site (str): HDX site to use eg. prod, test.
        hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
        hdx_key (str): Your HDX key. Ignored if hdx_read_only = True.
        hdx_config_dict (dict): HDX configuration dictionary to use instead of above 3 parameters OR
        hdx_config_json (str): Path to JSON HDX configuration OR
        hdx_config_yaml (str): Path to YAML HDX configuration
        project_config_dict (dict): Project configuration dictionary OR
        project_config_json (str): Path to JSON Project configuration OR
        project_config_yaml (str): Path to YAML Project configuration
        hdx_base_config_dict (dict): HDX base configuration dictionary OR
        hdx_base_config_json (str): Path to JSON HDX base configuration OR
        hdx_base_config_yaml (str): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.
    """

    _configuration = None
    home_folder = expanduser("~")
    default_hdx_base_config_yaml = script_dir_plus_file(
        "hdx_base_configuration.yml", ConfigurationError
    )
    default_hdx_config_yaml = join(home_folder, ".hdx_configuration.yml")

    prefix = f"HDXPythonLibrary/{__version__}"

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()

        self._session = None
        self._remoteckan = None
        self._emailer = None

        hdx_base_config_found = False
        hdx_base_config_dict = kwargs.get("hdx_base_config_dict")
        if hdx_base_config_dict:
            hdx_base_config_found = True
            logger.info("Loading HDX base configuration from dictionary")

        hdx_base_config_json = kwargs.get("hdx_base_config_json", "")
        if hdx_base_config_json:
            if hdx_base_config_found:
                raise ConfigurationError(
                    "More than one HDX base configuration given!"
                )
            hdx_base_config_found = True
            logger.info(
                f"Loading HDX base configuration from: {hdx_base_config_json}"
            )
            hdx_base_config_dict = load_json(hdx_base_config_json)

        hdx_base_config_yaml = kwargs.get("hdx_base_config_yaml", "")
        if hdx_base_config_found:
            if hdx_base_config_yaml:
                raise ConfigurationError(
                    "More than one HDX base configuration given!"
                )
        else:
            if not hdx_base_config_yaml:
                hdx_base_config_yaml = (
                    Configuration.default_hdx_base_config_yaml
                )
                logger.info(
                    f"No HDX base configuration parameter. Using default base configuration file: {hdx_base_config_yaml}."
                )
            logger.info(
                f"Loading HDX base configuration from: {hdx_base_config_yaml}"
            )
            hdx_base_config_dict = load_yaml(hdx_base_config_yaml)

        hdx_config_found = False
        hdx_config_dict = kwargs.get("hdx_config_dict")
        if hdx_config_dict:
            hdx_config_found = True
            logger.info("Loading HDX configuration from dictionary")

        hdx_config_json = kwargs.get("hdx_config_json", "")
        if hdx_config_json:
            if hdx_config_found:
                raise ConfigurationError(
                    "More than one HDX configuration given!"
                )
            hdx_config_found = True
            logger.info(f"Loading HDX configuration from: {hdx_config_json}")
            hdx_config_dict = load_json(hdx_config_json)

        hdx_config_yaml = kwargs.get("hdx_config_yaml", "")
        if hdx_config_found:
            if hdx_config_yaml:
                raise ConfigurationError(
                    "More than one HDX configuration given!"
                )
        else:
            if not hdx_config_yaml:
                hdx_config_yaml = Configuration.default_hdx_config_yaml
                if isfile(hdx_config_yaml):
                    logger.info(
                        f"No HDX configuration parameter. Using default configuration file: {hdx_config_yaml}."
                    )
                else:
                    logger.info(
                        f"No HDX configuration parameter and no configuration file at default path: {hdx_config_yaml}."
                    )
                    hdx_config_yaml = None
                    hdx_config_dict = dict()
            if hdx_config_yaml:
                logger.info(
                    f"Loading HDX configuration from: {hdx_config_yaml}"
                )
                hdx_config_dict = load_yaml(hdx_config_yaml)

        self.data = merge_two_dictionaries(
            hdx_base_config_dict, hdx_config_dict
        )

        project_config_found = False
        project_config_dict = kwargs.get("project_config_dict")
        if project_config_dict is not None:
            project_config_found = True
            logger.info("Loading project configuration from dictionary")

        project_config_json = kwargs.get("project_config_json", "")
        if project_config_json:
            if project_config_found:
                raise ConfigurationError(
                    "More than one project configuration given!"
                )
            project_config_found = True
            logger.info(
                f"Loading project configuration from: {project_config_json}"
            )
            project_config_dict = load_json(project_config_json)

        project_config_yaml = kwargs.get("project_config_yaml", "")
        if project_config_found:
            if project_config_yaml:
                raise ConfigurationError(
                    "More than one project configuration given!"
                )
        else:
            if project_config_yaml:
                logger.info(
                    f"Loading project configuration from: {project_config_yaml}"
                )
                project_config_dict = load_yaml(project_config_yaml)
            else:
                project_config_dict = dict()

        self.data = merge_two_dictionaries(
            hdx_base_config_dict, project_config_dict
        )

        ua = kwargs.get("full_agent")
        if ua:
            self.user_agent = ua
        else:
            try:
                self.user_agent = UserAgent.get(
                    prefix=Configuration.prefix, **kwargs
                )
            except UserAgentError:
                self.user_agent = UserAgent.get(
                    prefix=Configuration.prefix, **self.data
                )
        hdx_url = kwargs.get("hdx_url")
        if not hdx_url:
            hdx_url = self.data.get("hdx_url")
        if hdx_url:
            hdx_site = "custom"
            self.hdx_site = "hdx_custom_site"
            hdx_url = hdx_url.rstrip("/")
            self.data[self.hdx_site] = {"url": hdx_url}
        else:
            hdx_site = kwargs.get("hdx_site")
            if not hdx_site:
                hdx_site = self.data.get("hdx_site")
                if not hdx_site:
                    hdx_site = "stage"
            self.hdx_site = f"hdx_{hdx_site}_site"
            if self.hdx_site not in self.data:
                raise ConfigurationError(
                    f"{self.hdx_site} not defined in configuration!"
                )
        hdx_read_only = kwargs.get("hdx_read_only")
        if hdx_read_only is None:
            hdx_read_only = self.data.get("hdx_read_only")
            if hdx_read_only is None:
                hdx_read_only = False
        self.hdx_read_only = hdx_read_only
        logger.info(f"Read only access to HDX: {str(self.hdx_read_only)}")
        hdx_key_site = f"hdx_key_{hdx_site}"
        hdx_key = kwargs.get(hdx_key_site)
        if not hdx_key:
            hdx_key = self.data.get(hdx_key_site)
            if not hdx_key:
                hdx_key = kwargs.get("hdx_key")
                if not hdx_key:
                    hdx_key = self.data.get("hdx_key")
        if hdx_key:
            self.hdx_key = hdx_key
            return
        if not self.hdx_read_only:
            raise ConfigurationError(
                "No HDX API key supplied as a parameter or in configuration!"
            )

    def set_read_only(self, read_only: bool = True) -> None:
        """
        Set HDX read only flag

        Args:
            read_only (bool): Value to set HDX read only flag. Defaults to True.
        Returns:
            None

        """
        self.hdx_read_only = read_only

    def set_api_key(self, apikey: str) -> None:
        """
        Set HDX api key

        Args:
            apikey (str): Value to set api key.
        Returns:
            None

        """
        self.hdx_key = apikey

    def get_api_key(self) -> Optional[str]:
        """
        Return HDX api key or None if read only

        Returns:
            Optional[str]: HDX api key or None if read only

        """
        if self.hdx_read_only:
            return None
        return self.hdx_key

    def get_user_agent(self) -> str:
        """
        Return user agent

        Returns:
            str: User agent

        """
        return self.user_agent

    def get_hdx_site_url(self) -> str:
        """
        Return HDX web site url

        Returns:
            str: HDX web site url

        """
        return self.data[self.hdx_site]["url"]

    def get_dataset_url(self, name: str) -> str:
        """
        Return HDX dataset url given dataset name

        Args:
            name: Dataset name

        Returns:
            str: HDX dataset url

        """
        return f"{self.get_hdx_site_url()}/dataset/{name}"

    def _get_credentials(self) -> Optional[Tuple[str, str]]:
        """
        Return HDX site username and password

        Returns:
            Optional[Tuple[str, str]]: HDX site username and password or None

        """
        site = self.data[self.hdx_site]
        username = site.get("username")
        if username:
            return b64decode(username).decode("utf-8"), b64decode(
                site["password"]
            ).decode("utf-8")
        else:
            return None

    def get_session(self) -> requests.Session:
        """
        Return the session object

        Returns:
            requests.Session: The session object

        """
        if self._session is None:
            raise ConfigurationError(
                "There is no session set up! Use Configuration.create(**kwargs)"
            )
        return self._session

    def remoteckan(self) -> ckanapi.RemoteCKAN:
        """
        Return the remote CKAN object (see ckanapi library)

        Returns:
            ckanapi.RemoteCKAN: The remote CKAN object

        """
        if self._remoteckan is None:
            raise ConfigurationError(
                "There is no remote CKAN set up! Use Configuration.create(**kwargs)"
            )
        return self._remoteckan

    def call_remoteckan(self, *args: Any, **kwargs: Any) -> Dict:
        """
        Calls the remote CKAN

        Args:
            *args: Arguments to pass to remote CKAN call_action method
            **kwargs: Keyword arguments to pass to remote CKAN call_action method

        Returns:
            Dict: The response from the remote CKAN call_action method

        """
        requests_kwargs = kwargs.get("requests_kwargs", dict())
        credentials = self._get_credentials()
        if credentials:
            requests_kwargs["auth"] = credentials
        kwargs["requests_kwargs"] = requests_kwargs
        apikey = kwargs.get("apikey", self.get_api_key())
        kwargs["apikey"] = apikey
        return self.remoteckan().call_action(*args, **kwargs)

    @staticmethod
    def _environment_variables(**kwargs: Any) -> Any:
        """
        Overwrite keyword arguments with environment variables

        Args:
            **kwargs: See below
            hdx_url (str): HDX url to use. Overrides hdx_site.
            hdx_site (str): HDX site to use eg. prod, test. Defaults to test.
            hdx_key (str): Your HDX key. Ignored if hdx_read_only = True.

        Returns:
            kwargs: Changed keyword arguments

        """

        hdx_key = os.getenv("HDX_KEY")
        if hdx_key is not None:
            kwargs["hdx_key"] = hdx_key
        hdx_url = os.getenv("HDX_URL")
        if hdx_url is not None:
            kwargs["hdx_url"] = hdx_url
        else:
            hdx_site = os.getenv("HDX_SITE")
            if hdx_site is not None:
                kwargs["hdx_site"] = hdx_site
        return kwargs

    @classmethod
    def create_session_user_agent(
        cls,
        session: requests.Session = None,
        user_agent: Optional[str] = None,
        user_agent_config_yaml: Optional[str] = None,
        user_agent_lookup: Optional[str] = None,
        use_env: bool = False,
        **kwargs: Any,
    ) -> Tuple[requests.Session, str]:
        """
        Create session and user agent from configuration

        Args:
            session (requests.Session): requests Session object to use. Defaults to calling hdx.utilities.session.get_session()
            user_agent (Optional[str]): User agent string. HDXPythonLibrary/X.X.X- is prefixed.
            user_agent_config_yaml (Optional[str]): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (Optional[str]): Lookup key for YAML. Ignored if user_agent supplied.
            use_env (bool): Whether to read environment variables. Defaults to False.

        Returns:
            Tuple[requests.Session, str]: Tuple of (session, user agent)

        """
        if not session:
            whitelist = frozenset(
                ["HEAD", "TRACE", "GET", "POST", "PUT", "OPTIONS", "DELETE"]
            )
            session = get_session(
                user_agent,
                user_agent_config_yaml,
                user_agent_lookup,
                use_env,
                prefix=Configuration.prefix,
                method_whitelist=whitelist,
                **kwargs,
            )
            ua = session.headers["User-Agent"]
        else:
            ua = kwargs.get("full_agent")
            if not ua:
                ua = UserAgent.get(
                    user_agent,
                    user_agent_config_yaml,
                    user_agent_lookup,
                    prefix=Configuration.prefix,
                    **kwargs,
                )
        return session, ua

    def setup_session_remoteckan(
        self, remoteckan: Optional[ckanapi.RemoteCKAN] = None, **kwargs: Any
    ) -> None:
        """
        Set up remote CKAN from provided CKAN or by creating from configuration

        Args:
            remoteckan (Optional[ckanapi.RemoteCKAN]): CKAN instance. Defaults to setting one up from configuration.

        Returns:
            None

        """
        self._session, user_agent = self.create_session_user_agent(
            full_agent=self.get_user_agent(), **kwargs
        )
        if remoteckan is None:
            self._remoteckan = ckanapi.RemoteCKAN(
                self.get_hdx_site_url(),
                user_agent=user_agent,
                session=self._session,
            )
        else:
            self._remoteckan = remoteckan

    def emailer(self) -> Email:
        """
        Return the Email object (see :any:`Email`)

        Returns:
            Email: The email object

        """
        if self._emailer is None:
            raise ConfigurationError(
                "There is no emailer set up! Use setup_emailer(Any)"
            )
        return self._emailer

    def setup_emailer(self, **kwargs: Any) -> None:
        """
        Set up emailer. Parameters in dictionary or file (eg. yaml below):
        | connection_type: "ssl"   ("ssl" for smtp ssl or "lmtp", otherwise basic smtp is assumed)
        | host: "localhost"
        | port: 123
        | local_hostname: "mycomputer.fqdn.com"
        | timeout: 3
        | username: "user"
        | password: "pass"

        Args:
            **kwargs: See below
            email_config_dict (dict): HDX configuration dictionary OR
            email_config_json (str): Path to JSON HDX configuration OR
            email_config_yaml (str): Path to YAML HDX configuration. Defaults to ~/hdx_email_configuration.yml.

        Returns:
            None
        """
        self._emailer = Email(**kwargs)

    @classmethod
    def read(cls) -> "Configuration":
        """
        Read the HDX configuration

        Returns:
            Configuration: The HDX configuration

        """
        if cls._configuration is None:
            raise ConfigurationError(
                "There is no HDX configuration! Use Configuration.create(**kwargs)"
            )
        return cls._configuration

    @classmethod
    def setup(
        cls, configuration: Optional["Configuration"] = None, **kwargs: Any
    ) -> None:
        """
        Construct the HDX configuration

        Args:
            configuration (Optional[Configuration]): Configuration instance. Defaults to setting one up from passed arguments.
            **kwargs: See below
            user_agent (str): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
            user_agent_config_yaml (str): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (str): Lookup key for YAML. Ignored if user_agent supplied.
            hdx_url (str): HDX url to use. Overrides hdx_site.
            hdx_site (str): HDX site to use eg. prod, test.
            hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
            hdx_key (str): Your HDX key. Ignored if hdx_read_only = True.
            hdx_config_dict (dict): HDX configuration dictionary to use instead of above 3 parameters OR
            hdx_config_json (str): Path to JSON HDX configuration OR
            hdx_config_yaml (str): Path to YAML HDX configuration
            project_config_dict (dict): Project configuration dictionary OR
            project_config_json (str): Path to JSON Project configuration OR
            project_config_yaml (str): Path to YAML Project configuration
            hdx_base_config_dict (dict): HDX base configuration dictionary OR
            hdx_base_config_json (str): Path to JSON HDX base configuration OR
            hdx_base_config_yaml (str): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.

        Returns:
            None

        """
        if configuration is None:
            cls._configuration = Configuration(**kwargs)
        else:
            cls._configuration = configuration

    @classmethod
    def _create(
        cls,
        configuration: Optional["Configuration"] = None,
        remoteckan: Optional[ckanapi.RemoteCKAN] = None,
        **kwargs: Any,
    ) -> str:
        """
        Create HDX configuration

        Args:
            configuration (Optional[Configuration]): Configuration instance. Defaults to setting one up from passed arguments.
            remoteckan (Optional[ckanapi.RemoteCKAN]): CKAN instance. Defaults to setting one up from configuration.
            **kwargs: See below
            user_agent (str): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
            user_agent_config_yaml (str): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (str): Lookup key for YAML. Ignored if user_agent supplied.
            hdx_url (str): HDX url to use. Overrides hdx_site.
            hdx_site (str): HDX site to use eg. prod, test.
            hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
            hdx_key (str): Your HDX key. Ignored if hdx_read_only = True.
            hdx_config_dict (dict): HDX configuration dictionary to use instead of above 3 parameters OR
            hdx_config_json (str): Path to JSON HDX configuration OR
            hdx_config_yaml (str): Path to YAML HDX configuration
            project_config_dict (dict): Project configuration dictionary OR
            project_config_json (str): Path to JSON Project configuration OR
            project_config_yaml (str): Path to YAML Project configuration
            hdx_base_config_dict (dict): HDX base configuration dictionary OR
            hdx_base_config_json (str): Path to JSON HDX base configuration OR
            hdx_base_config_yaml (str): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.

        Returns:
            str: HDX site url

        """
        kwargs = cls._environment_variables(**kwargs)
        cls.setup(configuration, **kwargs)
        cls._configuration.setup_session_remoteckan(remoteckan, **kwargs)
        return cls._configuration.get_hdx_site_url()

    @classmethod
    def create(
        cls,
        configuration: Optional["Configuration"] = None,
        remoteckan: Optional[ckanapi.RemoteCKAN] = None,
        **kwargs: Any,
    ) -> str:
        """
        Create HDX configuration. Can only be called once (will raise an error if called more than once).

        Args:
            configuration (Optional[Configuration]): Configuration instance. Defaults to setting one up from passed arguments.
            remoteckan (Optional[ckanapi.RemoteCKAN]): CKAN instance. Defaults to setting one up from configuration.
            **kwargs: See below
            user_agent (str): User agent string. HDXPythonLibrary/X.X.X- is prefixed. Must be supplied if remoteckan is not.
            user_agent_config_yaml (str): Path to YAML user agent configuration. Ignored if user_agent supplied. Defaults to ~/.useragent.yml.
            user_agent_lookup (str): Lookup key for YAML. Ignored if user_agent supplied.
            hdx_url (str): HDX url to use. Overrides hdx_site.
            hdx_site (str): HDX site to use eg. prod, test.
            hdx_read_only (bool): Whether to access HDX in read only mode. Defaults to False.
            hdx_key (str): Your HDX key. Ignored if hdx_read_only = True.
            hdx_config_dict (dict): HDX configuration dictionary to use instead of above 3 parameters OR
            hdx_config_json (str): Path to JSON HDX configuration OR
            hdx_config_yaml (str): Path to YAML HDX configuration
            project_config_dict (dict): Project configuration dictionary OR
            project_config_json (str): Path to JSON Project configuration OR
            project_config_yaml (str): Path to YAML Project configuration
            hdx_base_config_dict (dict): HDX base configuration dictionary OR
            hdx_base_config_json (str): Path to JSON HDX base configuration OR
            hdx_base_config_yaml (str): Path to YAML HDX base configuration. Defaults to library's internal hdx_base_configuration.yml.

        Returns:
            str: HDX site url

        """
        if cls._configuration is not None:
            raise ConfigurationError("Configuration already created!")
        return cls._create(
            configuration=configuration, remoteckan=remoteckan, **kwargs
        )

    @classmethod
    def delete(cls) -> None:
        """
        Delete global HDX configuration.

        Returns:
            None

        """
        cls._configuration = None
